
from IPython.display import Markdown, display
from collections import defaultdict
import sys
import os
import json
import pandas as pd
from typing import Any, TypeVar, List, Tuple, Dict, Union
import requests
from requests.auth import HTTPBasicAuth
import textwrap
import re
import unicodedata
from tqdm.notebook import tqdm
from rapidfuzz import process, fuzz
from pathlib import Path
from collections.abc import Sequence, MutableSequence

from ..intervals import *

from ..genelist import GeneList
from ..coords import gene_coords_region
#from ..information import *

from ..utils import shelve_it
cache_dir = Path(os.path.dirname(__file__)).parent / 'data'


class NotFound(Exception):
    """
    Exception raised when a gene or other entity is not found.

    Exception : 
        Does nothing. Just a return value placeholder.
    """
    pass


from collections import defaultdict
import re

def list_assemblies() -> None:
    """
    Lists available genome assemblies from UCSC genome browser.
    """
    api_url = f'https://api.genome.ucsc.edu/list/ucscGenomes'
    response = requests.get(api_url)
    if not response.ok:
        response.raise_for_status()
    records = []
    for name, data in response.json()['ucscGenomes'].items():
        records.append((name.ljust(8), data['organism'].ljust(20), data['scientificName']))
    df = pd.DataFrame().from_records(records, columns=['Assemblies', 'Species', 'Latin name'])
    # df.groupby(['species', 'latin']).apply(lambda _df: (_df.species, _df.latin, _df.assembly))
    df = df.groupby(['Species', 'Latin name']).agg(lambda sr: ', '.join(sr)).reset_index()

    df = df.style.set_properties(**{'text-align': 'left'})
     
    df = df.set_table_styles(
        [dict(selector='th', props=[('text-align', 'left')])])
     
    display(df)




@shelve_it()
def _ensembl_id(name:str, species:str) -> str:
    species = '_'.join(species.lower().split())
    server = "https://rest.ensembl.org"
    ext = f"/xrefs/symbol/{species}/{name}?"
    r = requests.get(server+ext, headers={ "Content-Type" : "application/json"})
    if not r.ok:
      r.raise_for_status()
    decoded = r.json()
    ensembl_ids = [x['id'] for x in decoded if x['type'] == 'gene']
    if not len(ensembl_ids) == 1:
        raise NotFound
    return ensembl_ids[0]


def ensembl_id(name:Union[str, list], species:str) -> str:
    """
    Get ENSEMBL ID for some gene identifier

    Parameters
    ----------
    name : 
        Gene identifier
    species : 
        Species latin name, e.g. "Homo sapiens"

    Returns
    -------
    :
        ENSEMBL ID

    Raises
    ------
    [](`~geneinfo.NotFound`)
        Raises exception if no ENSEMBL ID can be found.
    """

    if isinstance(name, GeneList):
        name = list(name)
        
    if type(name) is str:
        return _ensembl_id(name, species)
    else:
        return [_ensembl_id(x, species) for x in name]


@shelve_it()
def ensembl2symbol(ensembl_id:str) -> str:
    """
    Converts ENSEMBL ID to gene HGCN gene symbol    

    Parameters
    ----------
    ensembl_id : 
        ENSEMBL ID

    Returns
    -------
    :
        HGCN gene symbol

    Raises
    ------
    [](`~geneinfo.NotFound`)
        Raises exception if no HGCN gene symbol can be found.
    """
    server = "https://rest.ensembl.org"
    ext = f"/xrefs/id/{ensembl_id}?"
    r = requests.get(server+ext, headers={ "Content-Type" : "application/json"})
    if not r.ok:
        r.raise_for_status()
    decoded = r.json()
    symbols = [x['display_id'] for x in decoded if x['dbname'] == 'HGNC']
    if not len(symbols) == 1:
        raise NotFound
    return symbols[0]


def assemblies(species:str) -> str:
    species = '_'.join(species.lower().split())
    server = "https://rest.ensembl.org"
    ext = f"/info/assembly/{species}?"
    r = requests.get(server+ext, headers={ "Content-Type" : "application/json"})
    if not r.ok:
        r.raise_for_status()
    decoded = r.json()
    genus, sp = species.split('_')

    markdown = (
        f"**{genus.capitalize()} {sp.lower()}:** "
         f"{decoded['default_coord_system_version']}. "
         f"Older: {','.join(decoded['coord_system_versions'])}"
    )
#    display(Markdown(markdown))
    return decoded['coord_system_versions']
    
    # symbols = [x['display_id'] for x in decoded if x['dbname'] == 'HGNC']
    # if not len(symbols) == 1:
    #     raise NotFound
    # return symbols[0]

# def assembly_info(assembly:str) -> str:
#     server = "https://rest.ensembl.org"
#     ext = f"/info/genomes/{assembly}?"
#     r = requests.get(server+ext, headers={ "Content-Type" : "application/json"})
#     if not r.ok:
#         r.raise_for_status()
#     decoded = r.json()
# #    display(Markdown(markdown))
#     return decoded


def hgcn_symbol(name:str|list) -> str:
    """
    Get HGCN gene symbol for some gene identifier

    Parameters
    ----------
    name : 
        Gene identifier or sequence of identifiers.

    Returns
    -------
    :
        HGCN gene symbol

    Raises
    ------
    [](`~geneinfo.NotFound`)
        Raises exception if no HGCN gene symbol can be found.
    """ 
    if isinstance(name, GeneList):
        name = list(name)
        
    if type(name) is list or type(name) is set:
        return [ensembl2symbol(ensembl_id(n, species='Homo sapiens')) for n in name]
    else:
        return ensembl2symbol(ensembl_id(name, species='Homo sapiens'))


@shelve_it()
def ensembl2ncbi(ensembl_id):
    """
    Converts ENSEMBL ID to gene NCBI ID

    Parameters
    ----------
    ensembl_id : 
        ENSEMBL ID

    Returns
    -------
    :
        NCBI ID

    Raises
    ------
    [](`~geneinfo.NotFound`)
        Raises exception if no NCBI ID can be found.
    """
    server = "https://rest.ensembl.org"
    ext = f"/xrefs/id/{ensembl_id}?"
    r = requests.get(server+ext, headers={ "Content-Type" : "application/json"})
    if not r.ok:
      r.raise_for_status()
    decoded = r.json()
    ids = [x['primary_id'] for x in decoded if x['dbname'] == 'EntrezGene']
    if not len(ids) == 1:
        raise NotFound
    return int(ids[0])


@shelve_it()
def mygene_get_gene_info(
    query, species='human', scopes='hgnc', 
    fields='symbol,alias,name,type_of_gene,summary,genomic_pos,genomic_pos_hg19'):

    api_url = f"https://mygene.info/v3/query?q={query}&scopes={scopes}&species={species}&fields={fields}"    
    response = requests.get(api_url)
    if not response.ok:
        response.raise_for_status()
    result = response.json()
    if 'hits' in result:
        for hit in result['hits']:
            if (type(query) is not int and hit['symbol'].upper() == query.upper()) or hit['_id'] == str(query):
                return hit
    print(f"Gene not found: {query}", file=sys.stderr)


# @shelve_it()
# def gene_coord(query: Union[str, List[str]], species, assembly:str=None, 
#                pos_list=False) -> dict:
#     """
#     Retrieves genome coordinates one or more genes.

#     Parameters
#     ----------
#     query : 
#         Gene symbol or list of gene symbols
#     species :  
#         Species, E.g 'homo_sapiens'.
#     assembly :  
#         Genome assembly, by default most recent.
#     pos_list :
#         Wether to instead return a list of (chrom, position, name) tuples.

#     Returns
#     -------
#     :
#         Dictionary with gene names as keys and (chrom, start, end, strand) tuples 
#         as values, or a list of (chrom, position, name) tuples.
#     """

#     coords = {}
#     batch_size = 100
#     for i in range(0, len(query), batch_size):
#         data = ensembl_get_gene_info_by_symbol(
#             query[i:i+batch_size], assembly=assembly, species=species)
#         for name, props in data.items():
#             chrom, start, end, strand = (
#                 props['seq_region_name'], props['start'], props['end'], props['strand']
#             )
#             if not chrom.lower().startswith('contig') \
#                     and not chrom.lower().startswith('scaffold'):
#                 chrom = 'chr'+chrom
#             if strand == -1:
#                 strand = '-'
#             elif strand == 1:
#                 strand = '+'
#             else:
#                 strand = None
    
#             coords[name] = (chrom, start, end, strand)

#     if pos_list:
#         annot = []
#         for (name, (chrom, start, end, strand)) in coords.items():
#             annot.append((chrom, int((start+end)/2), name))
#         return sorted(annot, key=lambda x: x[1:])

#     return coords


def gene_info(query: Union[str, List[str]], 
              scopes:str='hgnc') -> None:
    """
    Displays HTML formatted information about one or more human genes.

    Parameters
    ----------
    query : 
        Gene symbol or list of gene symbols
    scopes :  optional
        Scopes for information search, by default 'hgnc'
    """

    if isinstance(query, GeneList):
        query = list(query)
            
    if type(query) is not list:
        query = [query]
        
    for gene in query:

        for i in range(3):
            try:
                top_hit = mygene_get_gene_info(
                    gene, species='human', scopes=scopes,
                    fields='symbol,alias,name,type_of_gene,summary,genomic_pos,genomic_pos_hg19')
            except KeyError:
                continue
            else:
                break


        tmpl = "**Symbol:** **_{symbol}_** "

        if 'type_of_gene' in top_hit:
            tmpl += "({type_of_gene})"

        if 'alias' in top_hit:
            if type(top_hit['alias']) is str:
                top_hit['aliases'] = top_hit['alias']
            else:
                top_hit['aliases'] = ', '.join(top_hit['alias'])
            tmpl += " &nbsp; &nbsp; &nbsp; &nbsp; **Aliases:** {aliases}"
        tmpl += '  \n'

        if 'name' in top_hit:
            tmpl += '*{name}*  \n'

        if 'summary' in top_hit:
            tmpl += "**Summary:** {summary}  \n"

        if 'genomic_pos' in top_hit and 'genomic_pos_hg19' in top_hit:
            if type(top_hit['genomic_pos']) is list:
                top_hit['hg38'] = ', '.join(['{chr}:{start}-{end}'.format(**d) 
                                             for d in top_hit['genomic_pos']])
            else:
                top_hit['hg38'] = '{chr}:{start}-{end}'.format(
                    **top_hit['genomic_pos'])
            if type(top_hit['genomic_pos_hg19']) is list:
                top_hit['hg19'] = ', '.join(['{chr}:{start}-{end}'.format(**d) 
                                             for d in top_hit['genomic_pos_hg19']])
            else:
                top_hit['hg19'] = '{chr}:{start}-{end}'.format(
                    **top_hit['genomic_pos_hg19'])            
            tmpl += "**Human genomic position:** {hg38} (hg38), {hg19} (hg19)  \n"

        tmpl += "[Gene card](https://www.genecards.org/"
        tmpl += "cgi-bin/carddisp.pl?gene={symbol})  \n".format(**top_hit)

        tmpl += "\n\n ----"

        display(Markdown(tmpl.format(**top_hit)))


@shelve_it()
def _ensembl_get_features_region(chrom, window_start, window_end, 
                                 features=['gene', 'exon'], assembly=None, 
                                 species=None):
                                #  species='homo_sapiens'):
    if chrom.startswith('chr'):
        chrom = chrom[3:]
    window_start, window_end = int(window_start), int(window_end)
    genes = {}    
    for start in range(window_start, window_end, 500000):
        end = min(start+500000, window_end)
        param_str = ';'.join([f"feature={f}" for f in features])
        if assembly:
            url = f"https://{assembly.lower()}.rest.ensembl.org"
            api_url = f"{url}/overlap/region/{species}/{chrom}:{start}-{end}?{param_str}"
        else:
            url = "http://rest.ensembl.org"
            api_url = f"{url}/overlap/region/{species}/{chrom}:{start}-{end}?{param_str}"
        response = requests.get(api_url, headers={'content-type': 'application/json'})

        if not response.ok:
            response.raise_for_status()
        params = response.json()

        for gene in params:
            genes[gene['id']] = gene
            
    return genes


# @shelve_it()
# def ensembl_get_gene_info_by_symbol(symbols, assembly=None, species='homo_sapiens'):
def ensembl_get_gene_info_by_symbol(symbols, assembly=None, species=None):

    if assembly == 'hg38':
        assembly='GRCh38'
    if assembly == 'hg19':
        assembly='GRCh37'

    if isinstance(symbols, GeneList):
        symbols = list(symbols)

    if type(symbols) is not list:
        symbols = [symbols]

    if assembly:
        server = f"https://{assembly.lower()}.rest.ensembl.org"
    else:
        server = "https://rest.ensembl.org"
    ext = f"/lookup/symbol/{species}"
    headers={ "Content-Type" : "application/json", "Accept" : "application/json"}
    r = requests.post(server+ext, headers=headers, 
                      data=f'{{ "symbols": {json.dumps(symbols)} }}')
    if not r.ok:
        r.raise_for_status()
    return r.json()


def ensembl_get_genes_region(chrom, window_start, window_end, assembly=None, 
                             species=None):
                            #  species='homo_sapiens'):
    
    gene_info = _ensembl_get_features_region(
        chrom, window_start, window_end, features=['gene'], assembly=assembly, species=species)
    exon_info = _ensembl_get_features_region(
        chrom, window_start, window_end, features=['exon'], assembly=assembly, species=species)

    exons = defaultdict(list)
    for key, info in exon_info.items():
        exons[info['Parent']].append((info['start'], info['end']))

    for key in gene_info:
        gene_info[key]['exons'] = []
        if 'canonical_transcript' in gene_info[key]:
            transcript = gene_info[key]['canonical_transcript'].split('.')[0]
            if transcript in exons:
                gene_info[key]['exons'] = sorted(exons[transcript])

    return gene_info

def gene_info_region(chrom:str, window_start:int, window_end:int, 
                    #  assembly:str='GRCh38', 
                     assembly:str=None) -> None:
    """
    Displays HTML formatted information about genes in a chromosomal region.

    Parameters
    ----------
    chrom : 
        Chromosome identifier
    window_start : 
        Start of region
    window_end : 
        End of region (end base not included)
    assembly : 
        Genome assembly, e.g. 'hg38' or 'rheMac10'
    """
    for gene in gene_coords_region(chrom, window_start, window_end, assembly):    
        gene_info(gene[0])


    
def fetch_tracks(assembly):
    resp = requests.get(f"https://api.genome.ucsc.edu/list/tracks?genome={assembly}")
    resp.raise_for_status()
    tracks = resp.json()[assembly]
    tups = [(name, data['longLabel']) for name, data in tracks.items()]
    return sorted(tups, key=lambda t: t[0].upper())

def list_ucsc_tracks(assembly=None, label_wrap=80):
    search_ucsc_tracks(assembly=assembly,label_wrap=label_wrap)

def search_ucsc_tracks(*queries, assembly=None, label_wrap=80):

    assert assembly is not None
    
    def normalize(name):
        name = name.lower()
        name = unicodedata.normalize("NFKD", name).encode("ascii", "ignore").decode()
        name = re.sub(r"[^\w\s]", "", name)
        return name.strip()

    tracks = fetch_tracks(assembly=assembly)
    track_names, track_labels = zip(*tracks)

    if not queries:  
        entries = tracks #names, labels = track_names, track_labels
    else:        
        normalized_track_names = [normalize(n) for n in track_names]
        matches = defaultdict(float)
        for query in queries:
            for word in query.split():
                # match = process.extractOne(query_normalized, normalized_names, scorer=fuzz.WRatio)
                # match = process.extractOne(query_normalized, normalized_names, scorer=fuzz.WRatio, score_cutoff=90.0)
                search = process.extract(normalize(word), normalized_track_names, scorer=fuzz.WRatio, score_cutoff=80.0, limit=100)
                for name, score, index in search:
                    matches[(name, index)] += score
        
        sorted_hits = sorted([(v, k) for k, v in matches.items()], reverse=True)

        entries = []
        for score, (name, index) in sorted_hits:
            entries.append((track_names[index], track_labels[index]))

    if entries:
        ljust = max([len(e[0]) for e in entries]) + 2
        for name, label in entries:
            print(name.ljust(ljust) + '\n'.join(textwrap.wrap(label, width=label_wrap, subsequent_indent=' '*ljust)))

def get_ucsc_track(track_name, assembly, chrom=None, start=None, end=None):
    url = "https://api.genome.ucsc.edu/getData/track"
    params = {"genome": assembly, "track": track_name}
    if chrom is not None:
        assert chrom.startswith('chr')
        params['chrom'] = chrom
        if start is not None and end is not None:
            params['start'] = str(start)
            params['end'] = str(end)            
    response = requests.get(url, params=params)
    if response.ok:
    #    response.raise_for_status()
        try:
            track_data = response.json().get(track_name, [])
        except json.JSONDecodeError:
            print("Paramters does not represent a valid query", file=sys.stderr)
            return

        if type(track_data) is list:
            print("Track has heterogenous data records. Returning only attributes (columns) shared by all entries.", file=sys.stderr)
            
            shared_keys = list(set([k for d in track_data for k in d]))
            _track_data = defaultdict(list)
            for d in track_data:
                for k, v in d.items():
                    _track_data[k].append(v)
            track_data = _track_data
                
        try:
            return pd.DataFrame(track_data)
        except ValueError:
            raise        

def download_4dn(identifier, user_4dn, secret_4dn, dowload_dir=os.getcwd(), 
                 pgbar=False
                 ):
    
    def download_file(url, dowload_dir=dowload_dir):
        if not os.path.exists(dowload_dir):
            os.makedirs(dowload_dir)
        file_name = url.split('/')[-1]
        with requests.get(url, stream=True) as r:
            r.raise_for_status()
            with open(file_name, 'wb') as f:
                if pgbar:
                    pbar = tqdm(total=int(r.headers['Content-Length']))
                for chunk in r.iter_content(chunk_size=8192):
                    if chunk:
                        f.write(chunk)         
                        if pgbar:
                            pbar.update(len(chunk))

    url = f"https://data.4dnucleome.org/ga4gh/drs/v1/objects/{identifier}/access/https"
    response = requests.get(url, auth=HTTPBasicAuth(user_4dn, secret_4dn))
    if not response.ok:
        assert 0
    info = response.json()    
    download_file(info["url"])



from IPython.display import display, SVG
import io
import os
import pandas as pd
from collections.abc import Callable
from typing import Any, TypeVar, List, Tuple, Dict, Union

import requests

from ..intervals import *

from ..genelist import GeneList


def _get_string_ids(my_genes):
    string_api_url = "https://string-db.org/api"
    output_format = "tsv-no-header"
    method = "get_string_ids"
    params = {
        "identifiers" : "\r".join(my_genes), # your protein list
        "species" : 9606, # species NCBI identifier 
        "limit" : 1, # only one (best) identifier per input protein
        "echo_query" : 1, # see your input identifiers in the output
        "caller_identity" : "geneinfo" # your app name
    }
    request_url = "/".join([string_api_url, output_format, method])
    results = requests.post(request_url, data=params)
    if not results.ok:
        results.raise_for_status()    
    string_identifiers = []
    for line in results.text.strip().split("\n"):
        l = line.split("\t")
        input_identifier, string_identifier = l[0], l[2]
        string_identifiers.append(string_identifier)
    return string_identifiers


def show_string_network(my_genes:Union[list,str], nodes:int=10) -> None:
    """
    Display STRING network for a list of genes.

    Parameters
    ----------
    my_genes : 
        List of gene symbols
    nodes : 
        Number of nodes to show, by default 10
    """

    if not os.path.exists('geneinfo_cache'): os.makedirs('geneinfo_cache')

    if isinstance(my_genes, GeneList):
        my_genes = list(my_genes)

    if type(my_genes) is str:
        my_genes = list(my_genes)    
    string_identifiers = _get_string_ids(my_genes)
    string_api_url = "https://string-db.org/api"
    # string_api_url = "https://version-11-5.string-db.org/api"
    output_format = "svg"
    method = "network"
    request_url = "/".join([string_api_url, output_format, method])
    params = {
        "identifiers" : "\r".join(string_identifiers), # your proteins
        "species" : 9606, # species NCBI identifier 
        "add_white_nodes": nodes, # add 15 white nodes to my protein 
        "network_flavor": "confidence", # show confidence links
        "caller_identity" : "geneinfo" # your app name
    }
    response = requests.post(request_url, data=params)
    if not response.ok:
        response.raise_for_status()    
    file_name = "geneinfo_cache/network.svg"
    with open(file_name, 'wb') as fh:
        fh.write(response.content)
    display(SVG('geneinfo_cache/network.svg'))


def string_network_table(my_genes:list, nodes:int=10) -> pd.DataFrame:
    """
    Retrieves STRING network for a list of genes and returns it as a pandas.DataFrame.

    Parameters
    ----------
    my_genes : 
        List of gene symbols
    nodes : 
        Number of nodes to show, by default 10

    Returns
    -------
    :
        STRING network information for specified genes.
    """
    if isinstance(my_genes, GeneList):
        my_genes = list(my_genes)
    
    if type(my_genes) is str:
        my_genes = list(my_genes)
    string_api_url = "https://string-db.org/api"
    output_format = "tsv"
    method = "network"
    request_url = "/".join([string_api_url, output_format, method])
    params = {
        "identifiers" : "\r".join(my_genes), # your proteins
        "species" : 9606, # species NCBI identifier 
        "add_white_nodes": nodes, # add 15 white nodes to my protein 
        "network_flavor": "confidence", # show confidence links
        "caller_identity" : "geneinfo" # your app name
    }
    response = requests.post(request_url, data=params)
    if not response.ok:
        response.raise_for_status()    
    return pd.read_table(io.StringIO(response.content.decode()))




from collections import defaultdict
from math import log10
from collections.abc import Callable
from typing import Any, TypeVar, List, Tuple, Dict, Union

import matplotlib.axes
import matplotlib.pyplot as plt

from ..intervals import *
from ..coords import gene_coords_region


def _plot_gene(name, txstart, txend, exons, offset, line_width, 
               min_visible_width, font_size, ax, highlight=False, clip_on=True):

    color='black'

    line = ax.plot([txstart, txend], [offset, offset], color=color, 
                   linewidth=line_width/5, alpha=0.5)
    line[0].set_solid_capstyle('butt')
    for start, end in exons:
        end = max(start+min_visible_width, end)
        line = ax.plot([start, end], [offset, offset], linewidth=line_width, 
                       color=color)
        line[0].set_solid_capstyle('butt')
        
    if highlight is True:
        ax.text(txstart, offset-.5, name, horizontalalignment='right', 
                verticalalignment='center', fontsize=font_size, clip_on=clip_on,
                weight='bold', color='red')
    elif type(highlight) is dict:
        ax.text(txstart, offset-.5, name, horizontalalignment='right', 
                verticalalignment='center',
            fontsize=font_size, clip_on=clip_on, 
            **highlight)
    else:
        ax.text(txstart, offset-.5, name, horizontalalignment='right', 
                verticalalignment='center', fontsize=font_size, color=color, 
                clip_on=clip_on)


def gene_plot(chrom:str, start:str, end:str, assembly:str, highlight:List[Dict]=[], 
              hard_limits:bool=False, exact_exons:bool=False, figsize:tuple=None, 
              aspect:float=1, despine:bool=False, clip_on:bool=True, 
              gene_density:float=60, font_size:int=None, return_axes:int=1
              ) -> Union[matplotlib.axes.Axes, List[matplotlib.axes.Axes]]:
    """
    Plots gene ideograms for a chromosomal region and returns axes for 
    plotting along the same chromosome coordinates.

    Parameters
    ----------
    chrom : 
        Chromosome identifier
    start : 
        Start of region
    end : 
        End of region (end base not included)
    assembly : 
        Genome assembly identifier
    highlight : 
        List or dictionary of genes to highlight on gene plot (see Examples), 
        by default []
    hard_limits : 
        Whether to truncate plot in the middle of a gene, by default False so 
        that genes are fully plotted.
    exact_exons : 
        Whether to plot exon coordinates exactly, by default False so that exons 
        are plotted as a minimum width.
    figsize : 
        Figure size specified as a (width, height) tuple, by default None 
        honoring the default matplotlib settings.
    aspect : 
        Size of gene plot height relative to the total height of the other axes, 
        by default 1
    despine : 
        Whether to remove top and right frame borders, by default False
    clip_on : 
        Argument passed to axes.Text, by default True
    gene_density : 
        Controls the density of gene ideograms in the plot, by default 60
    font_size : 
        Gene label font size, by default None, in which case it is calculated 
        based on the region size.
    return_axes : 
        The number of vertically stacked axes to return for plotting over the 
        gene plot, by default 1

    Returns
    -------
    :
        A single axes or a list of axes for plotting data over the gene plot.

    Examples
    --------
    ```python
    import geneinfo.information as gi
    # Set email for Entrez queries
    gi.email('your@email.com')

    # Highlight a single gene
    ax = gene_plot('chr1', 1000000, 2000000, 'hg38', highlight='TP53')
    ax.scatter(chrom_coordinates, values)

    # Highlight multiple genes
    ax = gene_plot('chr1', 1000000, 2000000, 'hg38', highlight=['TP53', 'BRCA1'])
    ax.scatter(chrom_coordinates, values)

    # Highlight genes with custom styles
    ax = gene_plot('chr1', 1000000, 2000000, 'hg38', 
                    highlight={'TP53': {'color': 'blue', 'weight': 'bold'}})
    ax.scatter(chrom_coordinates, values)

    # Muli-gene highlight with custom styles
    ax = gene_plot('chr1', 1000000, 2000000, 'hg38', 
                    highlight={'TP53': {'color': 'blue', 'weight': 'bold'}, 
                               'BRCA1': {'color': 'red'}})
    ax.scatter(chrom_coordinates, values)

    # Multiple axes for plotting over gene plot
    axes = gene_plot('chr1', 1000000, 2000000, 'hg38', return_axes=2)
    ax1, ax2 = axes
    ax1.scatter(chrom_coordinates, values1)
    ax2.scatter(chrom_coordinates, values2)

    # Custom figure size and aspect ratio
    axes = gene_plot('chr1', 1000000, 2000000, 'hg38', figsize=(10, 4), aspect=0.5)
    ax1, ax2 = axes
    ax1.scatter(chrom_coordinates, values1)
    ax2.scatter(chrom_coordinates, values2)
    ```

    """

    fig, axes = plt.subplots(return_axes+1, 1, figsize=figsize, sharex='col', 
                             sharey='row', 
                             gridspec_kw={'height_ratios': 
                                          [1/return_axes]*return_axes + [aspect]})
    plt.subplots_adjust(wspace=0, hspace=0.15)

    genes = gene_coords_region(chrom, start, end, assembly)

    for i, (name, chrom, txstart, txend, transcripts) in enumerate(genes):
        if len(transcripts) > 1:
            exons = [exon for transcript in transcripts for exon in transcript]
            exons = collapse(sorted(exons))
        else:
            exons = transcripts[0]
        genes[i] = (name, chrom, txstart, txend, exons)

    line_width = max(6, int(50 / log10(end - start)))-2
    if font_size is None:
        font_size = max(6, int(50 / log10(end - start)))
    label_width = font_size * (end - start) / gene_density
    if exact_exons:
        min_visible_exon_width = 0
    else:
        min_visible_exon_width = (end - start) / 1000

    plotted_intervals = defaultdict(list)
    for name, chrom, txstart, txend, exons in genes:
        # exons = [e for t in exons for e in t]
        gene_interval = [txstart-label_width, txend]
        max_gene_rows = 1000
        for offset in range(1, max_gene_rows, 1):
            if not intersect([gene_interval], plotted_intervals[offset]) and \
                not intersect([gene_interval], plotted_intervals[offset-1]) and \
                not intersect([gene_interval], plotted_intervals[offset+1]) and \
                not intersect([gene_interval], plotted_intervals[offset-2]) and \
                not intersect([gene_interval], plotted_intervals[offset+2]) and \
                not intersect([gene_interval], plotted_intervals[offset-3]) and \
                not intersect([gene_interval], plotted_intervals[offset+3]):
                break
        if plotted_intervals[offset]:
            plotted_intervals[offset] = union(plotted_intervals[offset], 
                                              [gene_interval])
        else:
            plotted_intervals[offset] = [gene_interval]

        if type(highlight) is list or type(highlight) is set:
            hl = name in highlight
        elif type(highlight) is dict or type(highlight) is defaultdict:
            hl = highlight[name]
        else:
            hl = None

        _plot_gene(name, txstart, txend, exons, 
                  offset, line_width, min_visible_exon_width, font_size, 
                  highlight=hl,
                  ax=axes[-1], clip_on=clip_on)

    if plotted_intervals:
        offset = max(plotted_intervals.keys())
    else:
        offset = 1

    if hard_limits:
        axes[-1].set_xlim(start, end)
    else:
        s, e = axes[-1].get_xlim()
        axes[-1].set_xlim(min(s-label_width/2, start), max(e, end))

    axes[-1].set_ylim(-2, offset+2)
    axes[-1].get_yaxis().set_visible(False)
    axes[-1].invert_yaxis()
    axes[-1].spines['top'].set_visible(False)
    axes[-1].spines['right'].set_visible(False)
    axes[-1].spines['left'].set_visible(False)

    # TODO: add optional title explaining gene name styles

    for ax in axes[:-1]:
        ax.set_xlim(axes[-1].get_xlim())

        if despine:
            ax.spines['top'].set_visible(False)
            ax.spines['right'].set_visible(False)
    
    if return_axes == 1:
        return axes[0]
    else:
        return axes[:-1]




import click
import sys
import pandas as pd
import math
import subprocess
import logging
import os
import re
from datetime import datetime
from distutils.dir_util import copy_tree



##########################################################################
#
#                   Initialize globals
#
##########################################################################
startTime = datetime.now()

tmp_dir = '.tmp/'
log_dir = 'logs/'
chain_dir = 'chains/'
examples_dir = 'examples/'
# assume liftOver is in sys PATH
liftover_path = 'liftOver'  
#chain_dir = os.path.abspath('./chains')

# stores remapped positions for fast re-access
# key = chro_pos, value = [chro, pos, flat=mapped/unmapped]
remapped_list = {}

# stores processed files, used for restore progress
file_list = []
# stores failed files
failed_files = []

# Valid chromosome names
#valid_chro_names = ['chr'+str(i) for i in range(1,23)]
#valid_chro_names.append('chrX')
#valid_chro_names.append('chrY')

# the distance to next remapp position
step_size = 500

# the number positions to search
steps = 4000

# global beta
beta = 2

# if the header line is written
unmapped_logger_header = False


################### stat counters ###################
total_seg = 0
lifted_seg = 0
remapped_seg =0
rejected_seg = 0
unmapped_seg = 0

total_pro = 0
lifted_pro = 0
remapped_pro = 0
rejected_pro = 0
unmapped_pro = 0




# file = '/Volumes/arraymapMirror/arraymap/hg18/19197950/19197950_MB66_6332/segments.tab'
# ch = 'hg18ToHg19.over.chain'


# input_dir = '/Volumes/arraymapMirror/arraymap/hg18/GSE49'
# input_dir = '/Volumes/arraymapMirror/arraymap/hg18/GSE1755'
# output_dir = '/Users/bogao/DataFiles/hg19'
# segments_files = []
# probes_files = []


##########################################################################
#
#                   Utility functions
#
##########################################################################

# Map the unmapped positions to their nearest mappable positions
#
# Param:
# fin: path of the unmapped file generated by liftover
# chain: path of the chain file, should be same as used by liftover
# remap: the remapped_list
#
# Use global params:
# steps, step_size 
# the searching range is 100(bps)/step_stize * steps in both direction.
#
# Return:
# a list of lists with chro, new_pos, name 
# -1 in exception
#
# Note: unmappable positions will be returned with value 0
def solveUnmappables(fin, chain, remap):
    
    try:
        logger = logging.getLogger('liftover')
        
        # read in unmapped file
        df = pd.read_table(fin, sep='\t', comment='#', header=None, names=['chro','start','end','name'])
        df.loc[df.chro == 'chr23', 'chro'] = 'chrX'
        df.loc[df.chro == 'chr24', 'chro'] = 'chrY'
        # keep new coordinates
        positions = []
        # number of items
        num_pos = df.shape[0]
        counter = 0
#        cmd = [liftover_path, './tmp/remap.bed', chain, './tmp/remap_new.bed', 
#            './tmp/remap.unmapped']
        cmd = [liftover_path, os.path.join(tmp_dir, 'remap.bed'), chain, 
               os.path.join(tmp_dir,'remap_new.bed'), os.path.join(tmp_dir,'remap.unmapped')]
        
        
        # For each unmapped postion,
        # if it is in the remapped_list, get new position from the list
        # otherwise, gradually search along both sides of the chromesome,
        # until a mappable position is found
        # If nothing is mappable in 20M base range, assume that pos is unmappable.
        for i in range(num_pos):

            chro = df.iloc[i,0]
            start = df.iloc[i,1]
            name = df.iloc[i,3]
            new_pos = -1
            new_chro = 'NA'
            key = '{}_{}'.format(chro, start)

            # use buffered mapping if possible
            if key in remap:
                new_chro = remap[key][0]
                new_pos = remap[key][1]
                flag = remap[key][2]
                if flag == 'mapped':
                    counter += 1
                else:
                    logger.warning('Failed to convert (cached): ' + str([chro, start, name]))
            # do a stepwise mapping
            else:
                with open(os.path.join(tmp_dir, 'remap.bed'), 'w') as f:
                    for i in range(1,steps):
                        print('{}\t{}\t{}\t{}'.format(chro, start+i*step_size, start+i*step_size+1, name), file=f)
                        print('{}\t{}\t{}\t{}'.format(chro, start-i*step_size, start-i*step_size+1, name), file=f)

                return_info = subprocess.run(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
                
                # check running result
                if return_info.returncode != 0 :
                    logger.warning('Approximate conversion failed, cmd error: ' + str([chro, start, name]))
                elif os.path.getsize(os.path.join(tmp_dir,'remap_new.bed')) == 0 :
                    logger.warning('Failed to convert (new): ' + str([chro, start, name]))
                    remap[key] = [new_chro, new_pos, 'unmapped']
                # use the first mapping result
                else:
                    with open(os.path.join(tmp_dir, 'remap_new.bed'), 'r') as f:
                        next(f)
                        for line in f:
                            line = line.split('\t')
                            if len(line) > 1:
                                new_chro = line[0]
                                new_pos = int(line[1])
                            if new_chro == chro:
                                remap[key] = [new_chro, new_pos, 'mapped']
                                counter += 1
                                break




                        # while True:
                        #     line = f.readline()
                        #     line = line.split('\t')
                        #     if len(line) > 1:
                        #         new_chro = line[0]
                        #         new_pos = int(line[1])
                        #         if new_chro == chro:
                        #             remap[key] = [new_chro, new_pos, 'mapped']
                        #             counter += 1
                        #             break


            positions.append([new_chro, new_pos, name])
            
        logger.info('Approximate conversion: %i/%i positions.', counter, num_pos)
        return positions
    
    
    except Exception as e:
        logger.exception('Failure in approximate conversion: %s', fin)
        return -1















# Convert the genome coordinates in segments.tab to the specified the edition
# according to the provided chain file.
#
# Params:
# fin: the path of the input file
# chain: the path of the chain file
# remap: the remapped_list
#
# Return: 
# 0 or -1
#
def convertSegments(fin, fo, chain, remap, remap_flag=True, new_colnames = []):
    
    logger = logging.getLogger('liftover')
    logger.info('Processing segment:\t%s', fin)

    try:

        df = pd.read_table(fin, sep='\t', low_memory=False, keep_default_na=False)


        # save original column name
        original_colnames = df.columns.values.tolist()
        
        #Rename columns for processing
        df.rename(columns={df.columns[0]:'sample_id', df.columns[1]:'chromosome', df.columns[2]:'start', 
                           df.columns[3]:'stop'}, inplace=True)
       
        #Save column names for order restore after processing.
        col_names = df.columns
        

        #Drop NA
        df = df.dropna(axis=0, how='any', subset=['chromosome', 'start', 'stop'])
        chro_name = str( df.loc[0,'chromosome'] )
        if 'chr' not in chro_name:
            df['chr'] = 'chr' + df['chromosome'].astype(str)
        else:
            df['chr'] = df['chromosome'].astype(str)

        #Force positions to be integer
        df.start = df.start.astype(int)
        df.stop = df.stop.astype(int)

        #Generate new columns for processing
        df['name'] = df.index
        
        # update global counter
        global total_seg
        this_total = df.shape[0]
        total_seg += this_total

        #Create a file of start coordinates
        df_starts = df.loc[:,['chr','start','stop','name']]
        df_starts['stop'] = df_starts.start + 1
        df_starts.to_csv(os.path.join(tmp_dir,'starts.bed'), sep=' ', index=False, header=False)


        #Create a file of end coordinates
        df_ends = df.loc[:,['chr','start','stop','name']]
        df_ends['start'] = df_ends.stop - 1
        df_ends.to_csv(os.path.join(tmp_dir, 'ends.bed'), sep=' ', index=False, header=False)

    
    
        #Convert the start coordinates
#        cmd = [liftover_path , './tmp/starts.bed' , chain , './tmp/starts_new.bed' ,
#            './tmp/starts.unmapped']
        
        cmd = [liftover_path , os.path.join(tmp_dir, 'starts.bed'), chain, 
               os.path.join(tmp_dir, 'starts_new.bed') , os.path.join(tmp_dir, 'starts.unmapped')]
        
        return_info = subprocess.run(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        if return_info.returncode != 0 :
            logger.error('sh: %s', cmd)
            raise RuntimeError(cmd)
            
    
        #Read in the new start positions from a file
        starts_new = pd.read_table(os.path.join(tmp_dir, 'starts_new.bed'), sep='\t', names=df_starts.columns)
        del starts_new['stop']
        # update counter
#        lifted_start = starts_new.shape[0]
#        remapped_start = 0

        #Remap unmapped start positions
        if (remap_flag == True) and (os.path.getsize(os.path.join(tmp_dir, 'starts.unmapped')) >0):
            starts_remap = solveUnmappables(os.path.join(tmp_dir, 'starts.unmapped'), chain, remap)
            starts_remap = pd.DataFrame(starts_remap, columns=starts_new.columns)
            # update counter
#            remapped_start = starts_remap.shape[0]
            #Merge start positions
            starts_new = starts_new.append(starts_remap)
        else: 
            starts_remap = pd.DataFrame(columns=starts_new.columns)
       
        #Convert the end coordinates
#        cmd = [liftover_path , './tmp/ends.bed' , chain ,  './tmp/ends_new.bed' ,
#            './tmp/ends.unmapped' ]
        cmd = [liftover_path , os.path.join(tmp_dir, 'ends.bed') , chain ,
               os.path.join(tmp_dir, 'ends_new.bed') , os.path.join(tmp_dir, 'ends.unmapped' )]
        return_info = subprocess.run(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        if return_info.returncode != 0 :
            logger.error('sh: %s', cmd)
            raise RuntimeError(cmd)
    
        #Read in the new end positions from a file
        ends_new = pd.read_table(os.path.join(tmp_dir,'ends_new.bed'), sep='\t', names=df_ends.columns)
        del ends_new['start']
        # update counter
#        lifted_end = ends_new.shape[0]
#        remapped_end = 0
        #ends_new.rename(columns={'start':'stop'}, inplace=True)


        #Remap unmapped end positions
        if (remap_flag == True) and (os.path.getsize(os.path.join(tmp_dir, 'ends.unmapped')) >0):
            ends_remap = solveUnmappables(os.path.join(tmp_dir, 'ends.unmapped'), chain, remap)
            ends_remap = pd.DataFrame(ends_remap, columns=ends_new.columns)
            # update counter
#            remapped_end = ends_remap.shape[0]
            #Merge end positions
            ends_new = ends_new.append(ends_remap)
        else:
            ends_remap = pd.DataFrame(columns=ends_new.columns)
        
        

        #Merge new positions with original data 
        dd = pd.merge(starts_new,ends_new,how='inner', on=['name'], suffixes=['_s', '_e'])
        
        # update counter
        #lifted = lifted_start if lifted_start < lifted_end else lifted_end
        #remapped = remapped_start if remapped_start > remapped_end else remapped_end
#        global lifted_seg
#        #lifted_seg += lifted
##        unique_lifted = pd.merge(starts_new, ends_new, how='outer', on=['name'])
#        lifted_seg += dd.shape[0]
        
        
        df_new = pd.merge(dd, df, how='left', on=['name'],suffixes=['_new','_old'])
        #df_new.drop(['chr', 'name', 'start_old', 'stop_old'], axis=1, inplace=True)

        
        #Generate new columns for error checking
        df_new['chr_cmp'] = (df_new.chr_s == df_new.chr_e)
        df_new['pos_cmpRatio'] = (df_new.stop_new - df_new.start_new) / (df_new.stop_old - df_new.start_old)
        
        #Check bad liftovers
        df_mis = df_new[    ((df_new.start_new == -1) | (df_new.stop_new == -1)) |
                            (df_new.chr_cmp == False) | 
                            ((df_new.pos_cmpRatio < (1/beta) ) | (df_new.pos_cmpRatio > beta))]

        # update global counter
        global remapped_seg, rejected_seg, unmapped_seg, lifted_seg, unmapped_logger_header
        unmapped = df_mis[(df_mis.start_new == -1) | (df_mis.stop_new == -1)].shape[0]
#        remapped_seg = remapped_seg + remapped 
        unmapped_seg += unmapped
        rejected = df_mis.shape[0] - unmapped
        rejected_seg = rejected_seg + rejected
        if remap_flag == True:
            uniqe_remapped = pd.merge(starts_remap, ends_remap, how='outer', on=['name'])
            remapped_seg += uniqe_remapped.shape[0]
        else:
            uniqe_remapped = pd.DataFrame()
        #lifted_seg = lifted_seg + this_total - unmapped - uniqe_remapped.shape[0]
        lifted_seg = lifted_seg + this_total - unmapped - rejected
        #Invoke unmapped logger
        unmapped_logger = logging.getLogger('unmapped')
        if unmapped_logger_header == False:
            unmapped_logger.info('{}\t{}\t{}\t{}\t{}\t{}'.format('chromosome','start','end','same_chr/new_chr','length_ratio/new_pos','file'))
            unmapped_logger_header = True
        #logging unmapped positions
        for index, row in df_mis.iterrows():
            unmapped_logger.info('{}\t{}\t{}\t{}\t{:.4f}\t{}'.format(row['chr'],
                row['start_old'],row['stop_old'],row['chr_cmp'],row['pos_cmpRatio'],fin))
                
        
        #Rename and rearrange columns back to the original order
        df_new = df_new[~df_new.name.isin(df_mis.name)]
        df_new.rename(columns={'start_new':'start', 'stop_new':'stop'}, inplace=True)
        df_new = df_new[col_names]
        
        #restore column names
        if len(new_colnames) > 0:
            df_new.rename(columns={'sample_id':new_colnames[0], 'chromosome':new_colnames[1],
                                   'start':new_colnames[2], 'stop':new_colnames[3]}, inplace=True)
        else:
            df_new.columns = original_colnames
            
        
        os.makedirs(os.path.dirname(fo), exist_ok=True)
        # print(fo)
        #df_new.to_csv(fo, sep='\t', index=False, float_format='%.4f') 
        df_new.to_csv(fo, sep='\t', index=False, float_format='%.4f')         
        logger.info('Finished\n')
        progress_logger = logging.getLogger('progress')
        progress_logger.info(fin)
        return 0
    
    except Exception as e:
        logger.exception('Failure in segment: %s', fin)
        global failed_files
        failed_files.append(fin)
        return -1
        
    













# Convert the genome coordinates in CNprobes.tab to the specified the edition
# according to the provided chain file.
#
# Params:
# fin: the path of the input file
# chain: the path of the chain file
# remap: the remapped_list
#
# Return: 
# 0 or -1
#
def convertProbes(fin, fo, chain, remap, remap_flag=True, new_colnames=[]):
    
    logger = logging.getLogger('liftover')
    logger.info('Processing probe:\t%s', fin)
    
    try:
        #col_names=['probeID', 'chro', 'pos', 'value']
        
        
        #Special cases, some files do not have an ID column
        #Will generate ID for the new file.
#         with open(fin, 'r') as f:
#             line = f.readline().split('\t')
#             if len(line) < 4:
                
#                 logger.error('Unknown format: %s', fin)
#                 raise

        #df = pd.read_table(fin, sep='\t', header=0, names=col_names )
        
        
        df = pd.read_table(fin, sep='\t', low_memory=False, keep_default_na=False)
        if df.columns.size < 4:
            df.insert(0, 'probe_id', 'ID_' + df.index.astype(str))
            #df['probe_id'] = 'ID_' + df.index.astype(str)

        # save original column name
        original_colnames = df.columns.values.tolist()        
        
        df.rename(columns={df.columns[0]:'probe_id', df.columns[1]:'chromosome',
                           df.columns[2]:'position'}, inplace=True)        
        
        #Save column names for later restore.
        col_names = df.columns
        
        
        #Drop NA
        df = df.dropna(axis=0, how='any', subset=['chromosome', 'position'])
        chro_name = str( df.loc[0,'chromosome'] )
        if 'chr' not in chro_name:
            df['chr'] = 'chr' + df['chromosome'].astype(str)        
        else:
            df['chr'] = df['chromosome'].astype(str)


        #Force positions to be integer
        df.position = df.position.astype(int)

        #Generate new columns for processing
        df['name'] = df.index

        #Filter chromosome names
        
        # update counter
        global total_pro
        total_pro += df.shape[0]

        #Create a file of probe coordinates
        df_probes = df.loc[:,['chr','position']]
        df_probes['pos1'] = df_probes.position + 1
        df_probes['name'] = df_probes.index
        df_probes.to_csv(os.path.join(tmp_dir, 'probes.bed'), sep=' ', index=False, header=False)

    
        #Convert the probe coordinates
        cmd = [liftover_path , os.path.join(tmp_dir,'probes.bed') , chain ,
               os.path.join(tmp_dir,'probes_new.bed') , os.path.join(tmp_dir, 'probes.unmapped')]
        return_info = subprocess.run(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        if return_info.returncode != 0 :
            logger.error('sh: %s', cmd)
            raise RuntimeError(cmd)
            
    
        #Read in the new probe positions from a file
        probes_new = pd.read_table(os.path.join(tmp_dir,'probes_new.bed'), sep='\t', names=df_probes.columns)
        del probes_new['pos1']
        # update counter
        global lifted_pro
        lifted_pro += probes_new[probes_new.position !=-1].shape[0]
        remapped = 0


        #Remap the unmapped
        if (remap_flag == True) and (os.path.getsize(os.path.join(tmp_dir,'probes.unmapped')) >0):
            probes_remap = solveUnmappables(os.path.join(tmp_dir,'probes.unmapped'), chain, remap)
            probes_remap = pd.DataFrame(probes_remap, columns=probes_new.columns)
            # update counter
            remapped = probes_remap.shape[0]
            #Merage new positions
            probes_new = probes_new.append(probes_remap)
        
        #Merge and rearrange the coloumns to the original format
        df_new = pd.merge(probes_new, df, how='left', on=['name'],suffixes=['_new','_old'])
        
        #Check if new and old positions are on the chromosome
        df_new['chr_cmp'] = (df_new.chr_new == df_new.chr_old)        
        #Check if the new position is unmappable
        #Merge all unmapped positions
        df_mis = df_new[ (df_new.chr_cmp == False) | (df_new.position_new == -1)]
        
        # update global counter
        global remapped_pro, rejected_pro, unmapped_pro, unmapped_logger_header
        unmapped = df_mis[df_mis.position_new == -1].shape[0]
        unmapped_pro += unmapped
        remapped_pro = remapped_pro + remapped - unmapped
        rejected_pro = rejected_pro + df_mis.shape[0] - unmapped

        #Invoke unmapped logger
        unmapped_logger = logging.getLogger('unmapped')
        if unmapped_logger_header == False:
            unmapped_logger.info('{}\t{}\t{}\t{}\t{}\t{}'.format('chromosome','start','end','same_chr/new_chr','length_ratio/new_pos','file'))        
            unmapped_logger_header = True
        #logging unmapped positions
        for index, row in df_mis.iterrows():
            unmapped_logger.info('{}\t{}\t{}\t{}\t{}\t{}'.format( row['chr_old'],
                row['position_old'], '-1', row['chr_new'], row['position_new'], fin))
        
        
        df_new = df_new[~df_new.name.isin(df_mis.name)]
        df_new.rename(columns={'position_new':'position'}, inplace=True)
        df_new = df_new[col_names]
        
        #restore column names
        if len(new_colnames) > 0:
            df_new.rename(columns={'probe_id':new_colnames[0], 'chromosome':new_colnames[1],
                                   'position':new_colnames[2]}, inplace=True)
        else:
            df_new.columns = original_colnames
        
        os.makedirs(os.path.dirname(fo), exist_ok=True)
        df_new.to_csv(fo, sep='\t', index=False, float_format='%.4f') 
        
        logger.info('Finished\n')
        progress_logger = logging.getLogger('progress')
        progress_logger.info(fin)
        return 0
    
    except Exception as e:
        logger.exception('Failure in probe: %s', fin)
        global failed_files
        failed_files.append(fin)
        return -1
        






##########################################################################
#
#                   Command line interface
#
##########################################################################

@click.command()
@click.option('-i', '--input_dir', help='The directory to start processing.')
@click.option('-o', '--output_dir', help='The directory to write new files.')
@click.option('-c', '--chain_file', help='Specify the chain file name.')
@click.option('-si', '--segment_input_file', help='Specify the segment input file name.')
@click.option('-so', '--segment_output_file', help='Specify the segment output file name.')
@click.option('-pi', '--probe_input_file', help='Specify the probe input file name.')
@click.option('-po', '--probe_output_file', help='Specify the probe output file name.')
@click.option('-l', '--liftover', 'liftover_path_usr', type=str, help='Specify the location of the UCSC liftover program.')
@click.option('-t', '--test_mode', type=int, help='Only process a limited number of files.')
@click.option('-f', '--file_indexing', is_flag=True, help='Only generate the index file.')
@click.option('-x', '--index_file', type=click.File('r'), help='Specify an index file containing file paths.')
@click.option('-m', '--mapping_file', type=click.File('r'), help='Specify a pre-defined file of position mappings.')
@click.option('--step_size', 'step_size_usr', default=400, help='The step size of approximate conversion (in bases, default:400).')
@click.option('--range', 'search_range', default=10, help='The searching range of approximate conversion (in kilo bases, default:10).')
@click.option('--beta', 'beta_usr', type=click.FLOAT, help='Parameter in quality control.')
@click.option('--no_approximate_conversion', is_flag=True, help='Do not perform approximate conversion.')
@click.option('--new_segment_header', nargs=4, type=str, help='Specify 4 new column names for new segment files.' )
@click.option('--new_probe_header', nargs=3, type=str, help='Specify 3 new column names for new probe files.')
@click.option('--resume', 'resume_files', nargs=2, type=str, help='Specify a index file and a progress file to resume an interrupted job.')
@click.option('--demo', help='Copy example files to a user defined direcotry and run a demonstration.')
@click.option('--log_path', 'log_path_usr',type=str, help='Specify the directory to write logging files.')
def cli(input_dir, output_dir, chain_file, test_mode, file_indexing, segment_input_file, segment_output_file, 
        probe_input_file, probe_output_file, step_size_usr, search_range, index_file, mapping_file, no_approximate_conversion,
        new_segment_header, new_probe_header, resume_files, liftover_path_usr, beta_usr, demo, log_path_usr):


    test_counter = 0

    
    # User defined liftOver 
    if liftover_path_usr:
        global liftover_path
        liftover_path = liftover_path_usr
        
    # validate UCSC liftOver program
    try:
        subprocess.run(liftover_path,stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    except Exception as e:
        sys.exit('UCSC liftover program is not properly configurated: {}'.format(e))

    # Check beta value:
    global beta
    if beta_usr:
        beta = beta_usr
        if beta_usr <=0:
            sys.exit('Beta must be greater than zero.')        

    # run a demonstration
    # 1. copy examples files to user defined directory
    # 2. run with default params
    if demo:

        # check if user directory exists
        if os.path.isdir(demo) == False:
            sys.exit('Error: demo direcotry {} does not exist.'.format(demo))
        else:
            demo_input = os.path.join(os.path.abspath(demo), 'examples', 'inputs')

        # copy files
        examples_path = os.path.join(os.path.dirname(__file__), examples_dir)
        copy_tree(examples_path, demo_input, preserve_mode=0)
        if os.path.isdir(demo_input):
            print('Copied examples to {}'.format(demo_input))
        
        # default params
        input_dir = demo_input
        output_dir = os.path.join(os.path.abspath(demo), 'examples', 'outputs')
        chain_file = 'hg18ToHg38'
        segment_input_file = 'segments.tsv'
        segment_output_file = 'segments.tsv'
        probe_input_file = 'probes.tsv'
        probe_output_file = 'probes.tsv'
        print('Running a demonstration...\n')




    # Check params
#    if ((segment_input_file == None and segment_output_file !=None) or \
#        (segment_input_file != None and segment_output_file ==None)) :
#        sys.exit('Error: Must specify both input & output names for semgent or probe file.') 
#    if ((probe_input_file == None and probe_output_file != None) or \
#        (probe_input_file != None and probe_output_file == None)) :
#        sys.exit('Error: Must specify both input & output names for semgent or probe file.') 
#    if segment_input_file == None and segment_output_file ==None and \
#        probe_input_file == None and probe_output_file == None :
#        sys.exit('Error: Must specify input and output file names.')
    
    #check input & output file 
    if (segment_input_file == None) and (probe_input_file == None):
        sys.exit('Error: Must specify at least one input file name: semgent or probe.') 
        
    if segment_input_file:
        try:
            seg_pattern = re.compile(segment_input_file)
        except re.error:
            sys.exit('{} is not a valid regular expression.'.format(segment_input_file))
        # if segment_output_file == None:
        #     segment_output_file = segment_input_file
    # else:
    #     seg_pattern = re.compile('')
    
    if probe_input_file:
        try:
            pro_pattern = re.compile(probe_input_file)
        except re.error:
            sys.exit('{} is not a valid regular expression.'.format(probe_input_file))
        # if probe_output_file == None:
        #     probe_output_file = probe_input_file  
    # else:
    #     pro_pattern = re.compile('')



    # Validate input_dir and output_dir
    global tmp_dir, log_dir
    if input_dir:
        if os.path.isdir(input_dir) == False:
            sys.exit('Error: input direcotry does not exist.')
    else:
        sys.exit('Error: input_dir, out_dir, chain_file and a input file are required. Check --help for more information.')

    if output_dir == None:
        sys.exit('Error: input_dir, out_dir, chain_file and a input file are required. Check --help for more information.')
    else:
        tmp_dir = os.path.join(output_dir, tmp_dir)
        log_dir = os.path.join(output_dir, log_dir)
    
    if log_path_usr:
        if os.path.isdir(log_path_usr) == False:
            sys.exit("Error: log_path directory {} does not exist.".format(log_path_usr))
        else:
            log_dir = log_path_usr        
        
    # produce chain file
    default_chains = ['hg18ToHg19', 'hg18ToHg38', 'hg19ToHg38','hg38ToHg19','hg19ToHg18']
    if not chain_file:
        sys.exit('Error: please specify a chain file.')
    elif chain_file in default_chains:
        chain_file = chain_file + '.over.chain.gz'
        chain_file = os.path.join(os.path.dirname(__file__), chain_dir, chain_file )
    if os.path.isfile( chain_file ) == False:
        sys.exit('Error: chainfile does not exist.')


    # Assign step value
    global step_size, steps
    if step_size_usr > 0 :
        step_size = step_size_usr
    else:
        sys.exit('step_size must be greater than 0')
        
    if search_range >0 :
        steps = math.ceil(search_range*1000/step_size)
    else:
        sys.exit('range must be greater than 0')

    # convert no_approximate_conversion flg
    remap_flag = not no_approximate_conversion



    # create a directory for temp files, this dir is hard coded.
    os.makedirs(tmp_dir, exist_ok=True)
    
    ################### loggers ###################
    
    # check directory existance
    os.makedirs(log_dir, exist_ok=True)
    
    # system logger 
    logger = logging.getLogger('liftover')
    handler = logging.FileHandler(os.path.join(log_dir, 'general.log'), mode='w', delay=True)
    handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s'))
    logger.setLevel(logging.INFO)
    logger.addHandler(handler)
    
    # prgress logger, records processed files, used for restore.
    progress_logger = logging.getLogger('progress')
    handler = logging.FileHandler(os.path.join(log_dir,'progress.log'), mode='w', delay=True)
    handler.setFormatter(logging.Formatter('%(message)s'))
    progress_logger.setLevel(logging.INFO)
    progress_logger.addHandler(handler)
    
    # unmapped positions logger, records segments that's not properly lifted.
    unmapped_logger = logging.getLogger('unmapped')
    handler = logging.FileHandler(os.path.join(log_dir,'unconverted.log'), mode='w', delay=True)
    handler.setFormatter(logging.Formatter('%(message)s'))
    unmapped_logger.setLevel(logging.INFO)
    unmapped_logger.addHandler(handler)





    #########   Print input options   ############
    with open(os.path.join(log_dir,'parameters.log'), 'w') as fo:
        print('Command:', file=fo)
        print(*sys.argv, file=fo)
        print('Parameters:', file=fo )
        print('input_dir: {}'.format(input_dir), file=fo )
        print('output_dir: {}'.format(output_dir), file=fo )
        print('chain_file: {}'.format(chain_file), file=fo )
        print('test_mode: {}'.format(test_mode), file=fo )
        print('file_indexing: {}'.format(file_indexing), file=fo )
        print('segment_input_file: {}'.format(segment_input_file), file=fo )
        print('segment_output_file: {}'.format(segment_output_file), file=fo )
        print('probe_input_file: {}'.format(probe_input_file), file=fo )
        print('probe_output_file: {}'.format(probe_output_file), file=fo )
        print('setp_size: {}'.format(step_size_usr), file=fo )
        print('range: {}'.format( search_range ), file=fo )
        print('beta: {}'.format( beta ), file=fo)
        print('index_file: {}'.format( index_file.name if index_file else index_file ), file=fo )
        print('mapping_file: {}'.format( mapping_file.name if mapping_file else mapping_file), file=fo )
        print('no_approximate_conversion: {}'.format( no_approximate_conversion ), file=fo)
        print('new_segment_header: {}'.format( new_segment_header), file=fo)
        print('new_probe_header: {}'.format( new_probe_header), file=fo)
        print('log_path: {}'.format(log_dir), file=fo)
        print( file=fo)





    #########   Indexing all the files to be processed   ############

    #Recover the file list, if fileList.log exists.
    # if os.path.isfile(os.path.join(log_dir, 'fileList.log')):
    #     print('Index file detected, recovered from ./logs/fileList.log')
    #     with open(os.path.join(log_dir, 'fileList.log'), 'r') as fin:
    #         for line in fin:
    #             file_list.append(line.strip())

                # # test mode
                # if test_mode:
                #     test_counter += 1
                #     if test_counter > test_mode:
                #         break

    global file_list
    # resume function
    if len(resume_files) >0:
        if not os.path.isfile(resume_files[0]):
            sys.exit('--resume: index file does not exist.')
        if not os.path.isfile(resume_files[1]):
            sys.exit('--resume: progress file does not exist.')
        resume_index = []
        resume_progress = []
        with open(resume_files[0], 'r') as fi:
            for line in fi:
                resume_index.append(line.strip())
        with open(resume_files[1], 'r') as fi:
            for line in fi:
                resume_progress.append(line.strip())
        file_list = list(set(resume_index) - set(resume_progress))
        print('Resume from previous interruption, {} files processed, {} files to go.'.format(
                len(resume_progress), len(resume_index)-len(resume_progress)))


    elif index_file:
        for line in index_file:
            file_list.append(line.strip())
            # test mode
            if test_mode:
                test_counter += 1
                if test_counter >= test_mode:
                    break
        print('Index file detected, recovered from {}'.format(index_file.name))
        print('detected {} files.'.format(len(file_list)))
    #Traverse directories to index all segments.tab and CNprobes.tab files.
    else:
        print('Indexing files to process, this may take some time.')
        seg_counter = 0
        pro_counter = 0
        with click.progressbar(os.walk(input_dir), label='Be patient: ', fill_char=click.style('*', fg='red')) as bar:


            # File traverse
            for root, subdirs, files in bar:
                for f in files:
#                    if (f == segment_input_file):
                    if (segment_input_file !=None) and  (seg_pattern.match(f)):
                        path = os.path.join(root,f)
                        file_list.append(path)
                        seg_counter += 1
#                    elif(f == probe_input_file):
                    elif (probe_input_file !=None) and (pro_pattern.match(f)):
                        path = os.path.join(root,f)
                        file_list.append(path)
                        pro_counter += 1
                
                # test mode
                if test_mode:
                    test_counter += 1
                    if test_counter > test_mode:
                        break

            # Save the file list to disk
            with open(os.path.join(log_dir, 'fileList.log'), 'w') as fo:
                for line in file_list:
                    print(line,file=fo)

            # Terminate for file_indexing mode
            if file_indexing:
                sys.exit('Indexing file created.')
        print('detected {} segment files and {} probe files.\n'.format(seg_counter, pro_counter))




    # Recover the remapped_list if possible
    # if os.path.isfile('./logs/remapped.log'):
    #     with open('./logs/remapped.log', 'r') as fi:
    #         for line in fi:
    #             line = line.strip().split('\n')
    #             key = line[0]
    #             chro = line[1]
    #             pos = int(line[2])
    #             flag = line[3]
    #             remapped_list[key] = [chro, pos, flag]
    #     print('Remapped positions detected, recovered from ./logs/remapped.log')

    if mapping_file:
        if len(next(mapping_file).split('\t')) != 4:
            sys.exit('Wrong position mapping file.')
        for line in mapping_file:
            line = line.strip().split('\t')
            key = line[0]
            chro = line[1]
            pos = int(line[2])
            flag = line[3]
            remapped_list[key] = [chro, pos, flag]
        print('Position mapping file detected, recovered from {}'.format(mapping_file.name))

        



    # counters for display
    seg_succ_counter = 0
    seg_fail_counter = 0
    pro_succ_counter = 0
    pro_fail_counter = 0
    #########   Liftover   ############
    with click.progressbar(file_list, label='Lifting: ', fill_char=click.style('*', fg='green')) as bar:

        
        for f in bar:
            #generate output path
            rel_path = os.path.relpath(os.path.dirname(f), input_dir)
            #out_path = os.path.join(output_dir, rel_path, os.path.basename(f))


            
            # lift over
#            if os.path.basename(f) == segment_input_file:
            # if ((segment_input_file !=None) or (index_file != None)) and  (seg_pattern.match(os.path.basename(f))):
            if (segment_input_file !=None) and  (seg_pattern.match(os.path.basename(f))):
                if segment_output_file == None:
                    segment_output_file_dynamic = seg_pattern.match(os.path.basename(f)).group(0)
                    segment_out_path = os.path.join(output_dir, rel_path, segment_output_file_dynamic)
                else:
                    segment_out_path = os.path.join(output_dir, rel_path, segment_output_file)
                code = convertSegments(f, segment_out_path, chain_file,remapped_list, 
                                       remap_flag, new_segment_header)
                if code == 0:
                    seg_succ_counter += 1
                else:
                    seg_fail_counter += 1
#            elif os.path.basename(f) == probe_input_file:
            # elif ((probe_input_file !=None) or (index_file !=None )) and (pro_pattern.match(os.path.basename(f))):
            elif (probe_input_file !=None)  and (pro_pattern.match(os.path.basename(f))):
                if probe_output_file == None:
                    probe_output_file = pro_pattern.match(os.path.basename(f)).group(0)
                probe_out_path = os.path.join(output_dir, rel_path, probe_output_file)
                code = convertProbes(f, probe_out_path, chain_file, remapped_list, 
                                     remap_flag, new_probe_header)
                if code == 0:
                    pro_succ_counter += 1
                else:
                    pro_fail_counter += 1
            else:
                print('Unknown file type: ' + f)
                logger.error('Unknown file type: ' + f)
    
    if (seg_succ_counter + seg_fail_counter) >0:
        print('Segment files: {} processed, {} failed.'.format(seg_succ_counter, seg_fail_counter ))
    if (pro_succ_counter + pro_fail_counter) >0:
        print('Probe files: {} processed, {} failed'.format(pro_succ_counter, pro_fail_counter))


    # display global counts
    # print('Total segments: {}'.format(total_seg))
    # print('- directly converted: {}'.format(lifted_seg))
    # print('- approximately converted: {}'.format(remapped_seg))
    # print('- converted but rejected: {}'.format(rejected_seg))
    # print('- unconvertible: {}'.format(unmapped_seg))
    # print()
    # print('Total probes: {}'.format(total_pro))
    # print('- directly converted: {}'.format(lifted_pro))
    # print('- approximately converted: {}'.format(remapped_pro))
    # print('- converted but rejected: {}'.format(rejected_pro))
    # print('- unconvertible: {}'.format(unmapped_pro))   


    # Save the remapped_list for reuse
    with open(os.path.join(log_dir,'approximate_conversion.log'), 'w') as fo:
        print('{}\t{}\t{}\t{}'.format('name', 'new_chr', 'new_pos', 'result'), file=fo)
        for k,v in remapped_list.items():
            print(k, end='', file=fo)
            for i in v:
                print('\t{}'.format(i), end='', file=fo)
            print('', file=fo)
            
    # Save failed files
    if len(failed_files) >0:
        with open(os.path.join(log_dir,'failed_files.log'), 'w') as fo:
            for line in failed_files:
                print(line, file=fo)
    # Remove temp files.
#    subprocess.run('rm *.bed *.unmapped ./tmp/*.*  &>/dev/null', shell=True)
#    subprocess.run('rm -rf tmp', shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    print('Done! Finished in {}'.format(datetime.now() - startTime))

##########################################################################
#
#                   Main
#
##########################################################################
def main():    
    try:
        print()
        cli()
    except Exception as e:
        print(e)
    finally:
        subprocess.call(['rm','-rf',tmp_dir]) 
    
if __name__ == '__main__':
    main()



from collections import defaultdict
import sys
import os
import pandas as pd
from typing import Any, TypeVar, List, Tuple, Dict, Union
import requests
from pathlib import Path
from collections.abc import Sequence, MutableSequence

from ..utils import shelve_it, chrom_sort_key

cache_dir = Path(os.path.dirname(__file__)).parent / 'data'


def chromosome_lengths(assembly:str) -> List[Tuple[str, int]]:
    """
    Retrieves chromosome lengths for a genome assembly from UCSC.

    Parameters
    ----------
    assembly : 
        Assembly identifier, e.g. 'hg38' or 'rheMac10'

    Returns
    -------
    :
        List of tuples with chromosome name and length.
    """
    if assembly == 'hs1':
        return [
            ('chr1', 248387328),
            ('chr2', 242696752),
            ('chr3', 201105948),
            ('chr4', 193574945),
            ('chr5', 182045439),
            ('chr6', 172126628),
            ('chr7', 160567428),
            ('chr8', 146259331),
            ('chr9', 150617247),
            ('chr10', 134758134),
            ('chr11', 135127769),
            ('chr12', 133324548),
            ('chr13', 113566686),
            ('chr14', 101161492),
            ('chr15', 99753195),
            ('chr16', 96330374),
            ('chr17', 84276897),
            ('chr18', 80542538),
            ('chr19', 61707364),
            ('chr20', 66210255),
            ('chr21', 45090682),
            ('chr22', 51324926),
            ('chrX', 154259566),
            ('chrY', 62460029),
        ]

    api_url = f'https://api.genome.ucsc.edu/list/chromosomes?genome={assembly};track=gold'
    response = requests.get(api_url)
    if not response.ok:
        response.raise_for_status()
    records = []
    for chrom, length in response.json()['chromosomes'].items():
        if '_' in chrom or chrom == 'chrMT' or chrom == 'chrM':
            continue
        records.append((chrom, length))
    return sorted(records, key=chrom_sort_key)  


@shelve_it()
def centromere_coords(assembly:str) -> List[Tuple[str, int, int]]:
    """
    Retrieves centromere coordinates for a genome assembly from UCSC.

    Parameters
    ----------
    assembly : 
        Assembly identifier, e.g. 'hg38' or 'rheMac10'

    Returns
    -------
    :
        List of tuples with chromosome, start, and end of each centromere.
    """    

    if assembly == 'hs1':
        return [
            ('chr1', 121796048, 126300487),
            ('chr2', 92333543, 94673023),
            ('chr3', 91738002, 96415026),
            ('chr4', 49705154, 55199795),
            ('chr5', 47039134, 49596625),
            ('chr6', 58286706, 61058390),
            ('chr7', 60414372, 63714499),
            ('chr8', 44215832, 46325080),
            ('chr9', 44951775, 47582595),
            ('chr10', 39633793, 41664589),
            ('chr11', 51035789, 54450838),
            ('chr12', 34620838, 37202490),
            ('chr13', 15547593, 17498291),
            ('chr14', 10092112, 12708411),
            ('chr15', 16678794, 17694466),
            ('chr16', 35848286, 37829521),
            ('chr17', 23892419, 27486939),
            ('chr18', 15965699, 20933550),
            ('chr19', 25817676, 29768171),
            ('chr20', 26925852, 29099655),
            ('chr21', 10962853, 11306205),
            ('chr22', 12788180, 15711065),
            ('chrX', 57820107, 60927025),
            ('chrY', 10565750, 10883085),
        ]

    api_url = f'https://api.genome.ucsc.edu/getData/track?genome={assembly};track=centromeres'
    response = requests.get(api_url)
    if not response.ok:
        return []
    data = defaultdict(list)
    for chrom, val in response.json()['centromeres'].items():
        for d in val:
            if '_' in d['chrom'] or d['chrom'] == 'chrMT' or chrom == 'chrM':
                continue
            data[d['chrom']].append((d['chromStart'], d['chromEnd']))
    records = []
    for ch, val in data.items():
        starts, ends = zip(*val)
        records.append((ch, min(starts), max(ends)))
    return sorted(records, key=chrom_sort_key)  

    # df = pd.DataFrame().from_records(records, columns=['chrom', 'start', 'end'])
    # df.sort_values('chrom', key=lambda sr: [chrom_sort_key(x) for x in sr]).reset_index(drop=True)


@shelve_it()
def gene_coords_region(chrom=None, start=None, end=None, assembly=None, as_dataframe=False):
    """
    Gets gene structure information for genes in a chromosomal region.

    Parameters
    ----------
    chrom : 
        Chromosome identifier
    start : 
        Start of region
    end : 
        End of region (end base not included)
    assembly : 
        Genome assembly as USCS genome identifier. E.g. hg38 or rheMac10
    as_dataframe : 
        Return dataframe instead of list of tuples, by default False

    Returns
    -------
    :
        List of gene information. Each gene is a tuple with the following elements:
        - gene name
        - gene start
        - gene end
        - list of list of exons (start, end) for a transcript
    """    
    assert assembly is not None
    api_url = f'https://api.genome.ucsc.edu/getData/track?genome={assembly};track=ncbiRefSeq;chrom={chrom}'
    if start is not None and end is not None:
        api_url += f';start={start};end={end}'
    
    response = requests.get(api_url)
    if not response.ok:
        response.raise_for_status()

    data = defaultdict(list)
    for d in response.json()['ncbiRefSeq']:
        if d['chrom'] != chrom:
            continue
        exon_starts = [int(x) for x in d['exonStarts'].split(',') if x]
        exon_ends = [int(x) for x in d['exonEnds'].split(',') if x]
        exons = list(zip(exon_starts, exon_ends))    
        data[(d['chrom'], d['name2'])].append((d['txStart'], d['txEnd'], exons))
    records = []
    for (ch, name), val in data.items():
        starts, ends, exons = zip(*val)
        # records.append((chrom, name, starts, ends, exons))
        records.append((name, ch, min(starts), max(ends), exons))
    # for (chrom, name), (starts, ends, exons) in data.items():
    #     records.append((name, chrom, min(starts), max(ends), exons))
    if as_dataframe:
        return pd.DataFrame().from_records(records, columns=['name', 'chrom', 'start', 'end', 'exons'])
    else:
        return records


@shelve_it()
def all_gene_coords(assembly:str) -> Dict[str, Tuple[str, int, int]]:
    """
    Get gene coordinates for all genes in a genome assembly.

    Parameters
    ----------
    assembly : 
        Assembly identifier, e.g. 'hg38' or 'rheMac10'

    Returns
    -------
    :
        Dictionary where keys are gene names and each value is a tuple with chromosome, start, and end.
    """
    all_coords = {}
    for chrom, length in chromosome_lengths(assembly):
        if '_' in chrom or chrom == 'chrM':
            continue
        for tup in gene_coords_region(chrom, assembly=assembly, as_dataframe=True).itertuples():
            all_coords[tup.name] = (tup.chrom, tup.start, tup.end)
    return all_coords

    
@shelve_it()
def gene_coords(names:Union[str,Sequence[str]], assembly:str) -> List[Tuple[str, int, int, str]]:
    """
    Get gene coordinates for a gene or list of genes.

    Parameters
    ----------
    names : 
        Gene name or list of gene names
    assembly : 
        Assembly identifier, e.g. 'hg38' or 'rheMac10'

    Returns
    -------
    :
        List of tuples with chromosome, start, end, and gene name.
    """
    if type(names) is str:
        names = [names]
    all_genes = all_gene_coords(assembly)
    coords = []
    for name in names:
        try:
            coords.append((*all_genes[name], name))
        except KeyError:
            print(f"{name} not found", file=sys.stderr)
    return coords 


@shelve_it()
def gene_labels_region(chrom:str, start:int, end:int, assembly:str) -> List[Tuple[str, int, str]]:
    """
    Gets gene labels for genes in a chromosomal region. For use with the add_labels
    method of GenomeIdeogram and ChromIdeogram.

    Parameters
    ----------
    chrom : 
        Chromosome identifier, e.g. 'chr1'
    start : 
        Start coordinate of region
    end : 
        Start coordinate of region
    assembly : 
        Assembly identifier, e.g. 'hg38' or 'rheMac10'

    Returns
    -------
    :
        List of tuples with chromosome, position, and gene name.
    """
    df = gene_coords_region(chrom, start, end, assembly, as_dataframe=True)
    return list(zip(df.chrom, ((df.start+df.end)/2).astype(int), df.name))


@shelve_it()
def gene_labels(names:Sequence, assembly:str) -> List[Tuple[str, int, str]]:
    """
    Gets gene labels for a set of genes. For use with the add_labels
    method of GenomeIdeogram and ChromIdeogram.

    Parameters
    ----------
    names : 
        List of gene names.
    assembly : 
        Assembly identifier, e.g. 'hg38' or 'rheMac10'

    Returns
    -------
    :
        List of tuples with chromosome, position, and gene name.
    """
    if type(names) is str:
        names = [names]
    all_genes = all_gene_coords(assembly)
    coords = []
    for name in names:
        try:
            chrom, start, end = all_genes[name]
        except KeyError:
            print(f"{name} not found", file=sys.stderr)
        coords.append((chrom, (start + end)/2, name))
    return coords 


# @shelve_it()
# def get_genes_region(chrom:str, window_start:int, window_end:int, 
#                      assembly:str='GRCh38', db:str='ncbiRefSeq') -> list:
#     """
#     Gets gene structure information for genes in a chromosomal region.

#     Parameters
#     ----------
#     chrom : 
#         Chromosome identifier
#     window_start : 
#         Start of region
#     window_end : 
#         End of region (end base not included)
#     assembly : 
#         Genome assembly, by default 'GRCh38'
#     db : 
#         Database, by default 'ncbiRefSeq'

#     Returns
#     -------
#     :
#         List of gene information. Each gene is a tuple with the following elements:
#         - gene name
#         - gene start
#         - gene end
#         - gene strand
#         - list of exons (start, end)
#     """
#     api_url = f'https://api.genome.ucsc.edu/getData/track'
#     params = {'track': db,
#               'genome': assembly,
#               'chrom': chrom,
#               'start': window_start,
#               'end': window_end
#               }
#     response = requests.get(api_url, params=params)
#     if not response.ok:
#         response.raise_for_status()

#     genes = []
#     for gene in response.json()[db]:
#         exon_starts = [int(x) for x in gene['exonStarts'].split(',') if x]
#         exon_ends = [int(x) for x in gene['exonEnds'].split(',') if x]
#         exons = list(zip(exon_starts, exon_ends))
#         genes.append((gene['name2'], gene['txStart'], 
#                       gene['txEnd'], gene['strand'], exons))

#     return genes


# def get_genes_region_dataframe(chrom:str, window_start:int, window_end:int, 
#                      assembly:str=None, 
#                     #  assembly:str='GRCh38', 
#                      db:str='ncbiRefSeq') -> pd.DataFrame:
#     """
#     Gets gene structure information for genes in a chromosomal region in the form
#     of a pandas.DataFrame.

#     Parameters
#     ----------
#     chrom : 
#         Chromosome identifier
#     window_start : 
#         Start of region
#     window_end : 
#         End of region (end base not included)
#     assembly :  optional
#         Genome assembly, by default 'GRCh38'
#     db :  optional
#         Database, by default 'ncbiRefSeq'

#     Returns
#     -------
#     :
#         pandas.DataFrame with the following colunms:
#         - name: gene name
#         - start: gene start
#         - end: gene end
#         - strand: gene strand
#     """
#     try:
#         import pandas as pd
#     except ImportError:
#         print("pandas must be installed to return data frame")
#         return
#     genes = get_genes_region(chrom, window_start, window_end, assembly, db)
#     return pd.DataFrame().from_records([x[:4] for x in genes], 
#                                        columns=['name', 'start', 'end', 'strand'])








from IPython.display import Markdown, display, Image, SVG, HTML
from contextlib import redirect_stdout
import importlib
import sys
import os
import re
import subprocess
import pandas as pd
from collections.abc import Callable
from typing import Any, TypeVar, List, Tuple, Dict, Union

import matplotlib.axes
from matplotlib.patches import Rectangle, Polygon
import matplotlib.pyplot as plt

# from goatools.base import download_go_basic_obo
from goatools.obo_parser import GODag, OBOReader
from goatools.gosubdag.gosubdag import GoSubDag
from goatools.gosubdag.plot.gosubdag_plot import GoSubDagPlot
from goatools.cli.ncbi_gene_results_to_python import ncbi_tsv_to_py
from goatools.goea.go_enrichment_ns import GOEnrichmentStudyNS
from goatools.go_enrichment import GOEnrichmentRecord
from goatools.anno.genetogo_reader import Gene2GoReader
from goatools.go_search import GoSearch
from goatools.godag.go_tasks import CurNHigher
from goatools.godag_plot import (plot_gos, plot_goid2goobj, plot_results, 
                                 plt_goea_results)

from Bio import Entrez

from ..intervals import *

from ..information import hgcn_symbol, ensembl_id, NotFound


def email(email_address:str) -> None:
    """
    Registers your email address for Entrez queries. Thay way, NCBI will contect 
    you before closeing your connection if you are making too many queries.

    Parameters
    ----------
    email_address : 
        your email address
    """
    Entrez.email = email_address

def _assert_entrez_email():
    if not Entrez.email:
        print("""Please provide your email for Entrez queries:

import geneinfo.information as gi
gi.email("youremail@address.com)
""", file=sys.stderr)
        return

cache_dir = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'cache')

def download_ncbi_associations(prt=sys.stdout):

    # if not os.path.exists('geneinfo_cache'): os.makedirs('geneinfo_cache')
    if not os.path.exists(cache_dir): os.makedirs(cache_dir)

    if not os.path.exists(os.path.join(cache_dir, 'gene2go')):
        process = subprocess.Popen(['wget', '-nv', '-O', f'{cache_dir}/gene2go.gz',
                                    'https://ftp.ncbi.nlm.nih.gov/gene/DATA/gene2go.gz'],
                            stdout=subprocess.PIPE, 
                            stderr=subprocess.PIPE)
        stdout, stderr = process.communicate()
        print(stdout.decode(), file=prt)
        print(stderr.decode(), file=prt)
        assert not process.returncode

        process = subprocess.Popen(['gzip', '-f', '-d', f'{cache_dir}/gene2go.gz'],
                            stdout=subprocess.PIPE, 
                            stderr=subprocess.PIPE)
        stdout, stderr = process.communicate()
        print(stdout.decode(), file=prt)
        print(stderr.decode(), file=prt)    
        assert not process.returncode, process.returncode
    return f'{cache_dir}/gene2go'


def download_and_move_go_basic_obo(prt=sys.stdout):  

    if not os.path.exists(cache_dir): os.makedirs(cache_dir)

    if not os.path.exists(f'{cache_dir}/go-basic.obo'):
        # obo_fname = download_go_basic_obo(prt=prt)
        # shutil.move('go-basic.obo', 'geneinfo_cache/go-basic.obo')
        process = subprocess.Popen(['wget', '-nv', '-O', f'{cache_dir}/go-basic.obo',
                                    'https://purl.obolibrary.org/obo/go/go-basic.obo'],
                            stdout=subprocess.PIPE, 
                            stderr=subprocess.PIPE)
        stdout, stderr = process.communicate()
        print(stdout.decode(), file=prt)
        print(stderr.decode(), file=prt)
        assert not process.returncode

    return f'{cache_dir}/go-basic.obo'


def all_protein_coding(taxid=9606):
    return sorted(set(pd.read_csv(f'{cache_dir}/{taxid}_protein_genes.txt', sep='\t').Symbol))


def download_data(prt=sys.stdout):  

    download_ncbi_associations(prt)
    download_and_move_go_basic_obo(prt)


def _fetch_ids_to_file(id_list, output_file_name):

    with open(output_file_name, 'w') as f:
        header = ['tax_id', 'Org_name', 'GeneID', 'CurrentID', 'Status', 'Symbol', 
                  'Aliases', 'description', 'other_designations', 'map_location', 
                  'chromosome', 'genomic_nucleotide_accession.version', 
                  'start_position_on_the_genomic_accession', 
                  'end_position_on_the_genomic_accession', 'orientation', 
                  'exon_count', 'OMIM']
        print(*header, sep='\t', file=f)

        nr_genes_no_coordinates = 0

        batch_size = 2000
        for i in range(0, len(id_list), batch_size):
            to_fetch = id_list[i:i+batch_size]
            handle = Entrez.esummary(db="gene", id=",".join(to_fetch), 
                                     retmax=batch_size)
            entry = Entrez.read(handle)
            docsums = entry['DocumentSummarySet']['DocumentSummary']
            for doc in docsums:

                if doc['GenomicInfo']:
                    ver, start, stop, exon_count = (
                        doc['GenomicInfo'][0]['ChrAccVer'], 
                        doc['GenomicInfo'][0]['ChrStart'],
                        doc['GenomicInfo'][0]['ChrStop'],
                        doc['GenomicInfo'][0]['ExonCount'])
                else:
                    ver, start, stop, exon_count = 'unknown', pd.NA, pd.NA, pd.NA
                    nr_genes_no_coordinates += 1

                print(doc['Organism']['TaxID'], doc['Organism']['ScientificName'], 
                      doc.attributes['uid'], 
                      doc['CurrentID'], 
                      'live',
                      doc['Name'], doc['OtherAliases'], doc['Description'], 
                      doc['OtherDesignations'], doc['MapLocation'], 
                      doc['Chromosome'], ver, start, stop,
                      'notspecified', exon_count, '',
                      sep='\t', file=f)


def fetch_background_genes(taxid=9606):
    
    _assert_entrez_email()

    if not os.path.exists(cache_dir): os.makedirs(cache_dir)

    output_file_name = f'{cache_dir}/{taxid}_protein_genes.txt'   
    query = '[Taxonomy ID] AND alive[property] AND genetype protein coding[Properties]'     
    handle = Entrez.esearch(
        db="gene", 
        term=f'{taxid}{query}', 
        retmax="1000000")
    records = Entrez.read(handle)
    id_list = records["IdList"]

    _fetch_ids_to_file(id_list, output_file_name)

    # write mappings between symbol and ncbi id
    symbol2ncbi_file = f'{cache_dir}/{taxid}_symbol2ncbi.h5'
    df = pd.read_table(output_file_name)
    df = df.loc[:, ['GeneID', 'Symbol']]
    df.set_index('Symbol').GeneID.to_hdf(symbol2ncbi_file, key='symbol2ncbi')
    df.set_index('GeneID').Symbol.to_hdf(symbol2ncbi_file, key='ncbi2symbol')


def _cached_symbol2ncbi(symbols, taxid=9606):

    symbol2ncbi_file = f'{cache_dir}/{taxid}_symbol2ncbi.h5'
    symbol2ncbi = pd.read_hdf(symbol2ncbi_file, 'symbol2ncbi')
    try:    
        return symbol2ncbi.loc[symbols].tolist() 
    except KeyError:
        geneids = []
        for symbol in symbols:
            try:
                geneids.append(symbol2ncbi.loc[symbol])
            except KeyError:
                try:
                    ncbi_id = hgcn_symbol(symbol)
                    if ncbi_id not in symbol2ncbi.index:
                        print(ncbi_id, 'not in symbol2ncbi index')
                        raise NotFound
                        
                        
                    geneids.append(symbol2ncbi.loc[ncbi_id])
                    # geneids.append(ensembl2ncbi(ensembl_id(symbol)))                    
                except NotFound:
                    print(f'Could not map gene symbol "{symbol}" to ncbi id', 
                          file=sys.stderr)
        return geneids


def _cached_ncbi2symbol(geneids, taxid=9606):

    symbol2ncbi_file = f'{cache_dir}/{taxid}_symbol2ncbi.h5'
    symbol2ncbi = pd.read_hdf(symbol2ncbi_file, 'ncbi2symbol')
    try:
        return symbol2ncbi.loc[geneids].tolist()
    except KeyError:
        symbols = []
        for geneid in geneids:
            try:
                symbols.append(ncbi2symbol.loc[geneid])
            except KeyError:
                print(f'Could not map ncbi id "{geneid}" to gene symbol', 
                      file=sys.stderr)
        return symbols


def _tidy_taxid(taxid):
    try:
        taxid = int(taxid)
    except ValueError:
        handle = Entrez.esearch(db="taxonomy", term=f'"{taxid}"[Scientific Name]')
        id_list = Entrez.read(handle)['IdList']
        if id_list:
            taxid = int(id_list[0])
        else:
            print(f'Could not find taxonomy id for "{taxid}"')
    return taxid  
    

def symbols_protein_coding(taxid:int=9606) -> list:
    """
    List of protein coding gene symbols for a given taxonomy id.

    Parameters
    ----------
    taxid : 
        NCBI taxonomy ID, by default 9606 (which is human)

    Returns
    -------
    :
        List of gene symbols.
    """
    fetch_background_genes(taxid=taxid)
    symbol2ncbi_file = f'{cache_dir}/{taxid}_symbol2ncbi.h5'
    symbol2ncbi = pd.read_hdf(symbol2ncbi_file, 'ncbi2symbol')
    return symbol2ncbi.tolist()


def get_terms_for_go_regex(regex:str, taxid:int=9606, add_children:bool=False) -> list:
    """
    Get GO terms for terms matching a regular expression in their description string.

    Parameters
    ----------
    regex : 
        Regular expression to match GO term descriptions.
    taxid : 
        NCBI taxonomy ID, by default 9606 (which is human)
    add_children : 
        Add GO terms nested under GO terms found, by default False

    Returns
    -------
    :
        List of GO terms.
    """

    taxid = _tidy_taxid(taxid)
        
    with open(os.devnull, 'w') as null, redirect_stdout(null):

        obo_fname = download_and_move_go_basic_obo(prt=null)

        gene2go = download_ncbi_associations(prt=null)

        objanno = Gene2GoReader(f"{cache_dir}/gene2go", taxids=[taxid], prt=null)
        go2geneids = objanno.get_id2gos(namespace='*', go2geneids=True, prt=null)
        srchhelp = GoSearch(f"{cache_dir}/go-basic.obo", 
                            go2items=go2geneids, log=null)

        results_all = re.compile(r'({})'.format(regex), flags=re.IGNORECASE)
        results_not = re.compile(r'({}).independent'.format(regex), flags=re.IGNORECASE)

        gos_all = srchhelp.get_matching_gos(results_all, prt=null)
        gos_no = srchhelp.get_matching_gos(results_not, gos=gos_all)
        gos = gos_all.difference(gos_no)
        if add_children:
            gos = srchhelp.add_children_gos(gos)

        return list(gos)


def get_genes_for_go_regex(regex:str, taxid:int=9606) -> pd.DataFrame:
    """
    Get gene information for GO terms matching a regular expression in 
    their description string.

    Parameters
    ----------
    regex : 
        Regular expression to match GO term descriptions.
    taxid : 
        NCBI taxonomy ID, by default 9606 (which is human)

    Returns
    -------
    :
        Columns: symbol, name, chrom, start, end.
    """
    _assert_entrez_email()

    taxid = _tidy_taxid(taxid)
 
    with open(os.devnull, 'w') as null, redirect_stdout(null):

        gos_all_with_children = get_terms_for_go_regex(
            regex, taxid=taxid, add_children=True)

        objanno = Gene2GoReader(f"{cache_dir}/gene2go", taxids=[taxid], prt=null)
        go2geneids = objanno.get_id2gos(namespace='*', go2geneids=True, prt=null)
        srchhelp = GoSearch(f"{cache_dir}/go-basic.obo", go2items=go2geneids, 
                            log=null)
        geneids = srchhelp.get_items(gos_all_with_children)

        ncbi_tsv = f'{cache_dir}/{taxid}_protein_genes.txt'
        if not os.path.exists(ncbi_tsv):
            fetch_background_genes(taxid)

        output_py = f'{cache_dir}/{taxid}_protein_genes.py'
        ncbi_tsv_to_py(ncbi_tsv, output_py, prt=null)
        
        protein_genes = importlib.import_module(
            output_py.replace('.py', '').replace('/', '.'))
        GENEID2NT = protein_genes.GENEID2NT

    fetch_ids = geneids

    fetch_ids = list(map(str, fetch_ids))
    records = []
    found = []
    batch_size = 2000
    for i in range(0, len(fetch_ids), batch_size):
        to_fetch = fetch_ids[i:i+batch_size]
        handle = Entrez.esummary(db="gene", id=",".join(to_fetch), retmax=batch_size)
        entry = Entrez.read(handle)
        docsums = entry['DocumentSummarySet']['DocumentSummary']
        for doc in docsums:
            try:
                chrom_pos = (doc['Chromosome'], doc['GenomicInfo'][0]['ChrStart'],
                              doc['GenomicInfo'][0]['ChrStop'])
            except:
                print(f"WARNING: no coordinates for {doc['Name']} (pandas.NA)", 
                      file=sys.stderr)
                chrom_pos = (pd.NA, pd.NA, pd.NA)
            records.append((doc['Name'], doc['Description'], *chrom_pos))
            found.append(str(doc.attributes['uid']))
    missing = set(fetch_ids).difference(set(found))

    df = pd.DataFrame().from_records(
        records, columns=['symbol', 'name', 'chrom', 'start', 'end'])

    return df.sort_values(by='start').reset_index(drop=True)

    
def get_genes_for_go_terms(terms, taxid=9606) -> pd.DataFrame:
    """
    Get gene information for genes with specified GO terms.

    Parameters
    ----------
    terms : 
        List of GO terms
    taxid : 
        NCBI taxonomy ID, by default 9606 (which is human)

    Returns
    -------
    :
        Columns: symbol, name, chrom, start, end.
    """

    if type(terms) is not list:
        terms = [terms]

    with open(os.devnull, 'w') as null, redirect_stdout(null):

        obo_fname = download_and_move_go_basic_obo(prt=null)
        gene2go = download_ncbi_associations(prt=null)
        objanno = Gene2GoReader(f"{cache_dir}/gene2go", taxids=[taxid], prt=null)
        go2geneids = objanno.get_id2gos(namespace='*', go2geneids=True, prt=null)
        srchhelp = GoSearch(f"{cache_dir}/go-basic.obo", go2items=go2geneids, 
                            log=null)

        geneids = srchhelp.get_items(terms)  

        ncbi_tsv = f'{cache_dir}/{taxid}_protein_genes.txt' 
        if not os.path.exists(ncbi_tsv):
            fetch_background_genes(taxid)

        output_py = f'{cache_dir}/{taxid}_protein_genes.py'
        ncbi_tsv_to_py(ncbi_tsv, output_py, prt=null)

    protein_genes = importlib.import_module(
        output_py.replace('.py', '').replace('/', '.'))
    GENEID2NT = protein_genes.GENEID2NT

    fetch_ids = geneids

    fetch_ids = list(map(str, fetch_ids))
    records = []
    found = []
    batch_size = 2000
    for i in range(0, len(fetch_ids), batch_size):
        to_fetch = fetch_ids[i:i+batch_size]
        handle = Entrez.esummary(db="gene", id=",".join(to_fetch), retmax=batch_size)
        entry = Entrez.read(handle)
        docsums = entry['DocumentSummarySet']['DocumentSummary']
        for doc in docsums:
            try:
                chrom_pos = (doc['Chromosome'], doc['GenomicInfo'][0]['ChrStart'], 
                             doc['GenomicInfo'][0]['ChrStop'])
            except:
                print(f"WARNING: no coordinates for {doc['Name']} (pandas.NA)", 
                      file=sys.stderr)
                chrom_pos = (pd.NA, pd.NA, pd.NA)
            records.append((doc['Name'], doc['Description'], *chrom_pos))
            found.append(str(doc.attributes['uid']))
    missing = set(fetch_ids).difference(set(found))

    df = pd.DataFrame().from_records(
        records, columns=['symbol', 'name', 'chrom', 'start', 'end'])

    return df.sort_values(by='start').reset_index(drop=True)


def go_annotation_table(taxid:int=9606) -> pd.DataFrame:
    """
    GO annotations for a given taxonomy id as a pandas.DataFrame.

    Parameters
    ----------
    taxid : 
        NCBI taxonomy ID, by default 9606, which is human

    Returns
    -------
    pd.DataFrame
        GO annotations for the specified taxonomy id.
    """
    _assert_entrez_email()

    try:
        taxid = int(taxid)
    except ValueError:
        handle = Entrez.esearch(db="taxonomy", term=f'"{taxid}"[Scientific Name]')
        taxid = int(Entrez.read(handle)['IdList'][0])
        
    with open(os.devnull, 'w') as null, redirect_stdout(null):

        obo_fname = download_and_move_go_basic_obo(prt=null)

        gene2go = download_ncbi_associations(prt=null)
        
    df = pd.read_table(gene2go, sep='\t')
    df.rename(columns={'#tax_id': 'taxid'}, inplace=True)
    return df.loc[df['taxid'] == taxid]


def gene_annotation_table(taxid:int=9606) -> pd.DataFrame:
    """
    Gene annotations for a given taxonomy id as a pandas.DataFrame.

    Parameters
    ----------
    taxid : 
        NCBI taxonomy ID, by default 9606, which is human

    Returns
    -------
    pd.DataFrame
        Gene annotations for the specified taxonomy id.
    """

    ncbi_tsv = f'{cache_dir}/{taxid}_protein_genes.txt'
    if not os.path.exists(ncbi_tsv):
        fetch_background_genes(taxid)
    df = pd.read_table(ncbi_tsv)
    df.rename(columns={'tax_id': 'taxid'}, inplace=True)
    return df.loc[df['taxid'] == taxid]


def get_go_terms_for_genes(genes:Union[str,list], taxid:int=9606, 
                           evidence:list=None) -> list:
    """
    Get the union of GO terms for a list of genes.

    Parameters
    ----------
    genes : 
        Gene name or list of gene names.
    taxid : 
        NCBI taxonomy ID, by default 9606, which is human.
    evidence : 
        Evidence codes, by default None

    Returns
    -------
    :
        Go terms for the specified genes.
    """
    go_df = go_annotation_table(taxid)
    genes_df = gene_annotation_table(taxid)
    gene_ids = genes_df.loc[genes_df.Symbol.isin(genes)].GeneID

    df = go_df.loc[go_df.GeneID.isin(gene_ids)]
    if len(df.index) and evidence is not None:
        df = df.loc[df.Evidence.isin(evidence)]
        
    return list(sorted(df.GO_ID.unique().tolist()))

    
def show_go_dag_for_terms(terms:Union[list, pd.Series], 
                          add_relationships:bool=True) -> None:
    """
    Display GO graph for a list of GO terms.

    Parameters
    ----------
    terms : 
        Go terms
    add_relationships : 
        Add edges representing relationships between GO terms, by default True
    """
    if type(terms) is pd.core.series.Series:
        terms = terms.tolist()

    if not terms:
        return

    with open(os.devnull, 'w') as null, redirect_stdout(null):

        obo_fname = download_and_move_go_basic_obo(prt=null)
        
        file_gene2go = download_ncbi_associations(prt=null)

        if add_relationships:
            optional_attrs=['relationship', 'def']
        else:
            optional_attrs=['def']
        obodag = GODag(f"{cache_dir}/go-basic.obo", optional_attrs=optional_attrs, 
                       prt=null)

        gosubdag = GoSubDag(terms, obodag, relationships=add_relationships) 
        GoSubDagPlot(gosubdag).plt_dag(f'{cache_dir}/plot.png')

    return display(Image(f'{cache_dir}/plot.png'))

# def show_go_dag_for_terms(terms, add_relationships=True):

#     with open(os.devnull, 'w') as null, redirect_stdout(null):
#         if add_relationships:
#             optional_attrs=['relationship', 'def']
#         else:
#             optional_attrs=['def']
#         obodag = GODag("go-basic.obo", optional_attrs=optional_attrs, prt=null)
#         plot_gos('plot.png', terms, obodag)
#     return Image('plot.png')  


# https://github.com/tanghaibao/goatools/
# blob/main/notebooks/goea_nbt3102_group_results.ipynb


def show_go_dag_for_gene(gene:str, taxid:int=9606, evidence:list=None, 
                         add_relationships:bool=True) -> None:
    """
    Displays GO graph for a given gene.

    Parameters
    ----------
    gene : 
        Gene symbol
    taxid : 
        NCBI taxonomy ID, by default 9606, which is human
    evidence : 
        Limiting list of evidence categories to include, by default None. 
        See `show_go_evidence_codes()`.
    add_relationships : 
        Add edges representing relationships between GO terms, by default True
    """

    # evidence codes: http://geneontology.org/docs/guide-go-evidence-codes/
    go_terms = get_go_terms_for_genes([gene], taxid=taxid, evidence=evidence)
    if not go_terms:
        print('No GO terms to show', file=sys.stderr)
        return
    return show_go_dag_for_terms(go_terms, add_relationships=add_relationships)

    
def show_go_evidence_codes() -> None:   
    """
    Display list of GO evidence categories and their codes.
    """

    s = """
**Experimental evidence codes:** <br>
Inferred from Experiment (EXP) <br>
Inferred from Direct Assay (IDA) <br>
Inferred from Physical Interaction (IPI) <br>
Inferred from Mutant Phenotype (IMP) <br>
Inferred from Genetic Interaction (IGI) <br>
Inferred from Expression Pattern (IEP) <br>
Inferred from High Throughput Experiment (HTP) <br>
Inferred from High Throughput Direct Assay (HDA) <br>
Inferred from High Throughput Mutant Phenotype (HMP) <br>
Inferred from High Throughput Genetic Interaction (HGI) <br>
Inferred from High Throughput Expression Pattern (HEP) 

**Phylogenetically-inferred annotations:** <br>
Inferred from Biological aspect of Ancestor (IBA) <br>
Inferred from Biological aspect of Descendant (IBD) <br>
Inferred from Key Residues (IKR) <br>
Inferred from Rapid Divergence (IRD)

**Computational analysis evidence codes** <br>
Inferred from Sequence or structural Similarity (ISS) <br>
Inferred from Sequence Orthology (ISO) <br>
Inferred from Sequence Alignment (ISA) <br>
Inferred from Sequence Model (ISM) <br>
Inferred from Genomic Context (IGC) <br>
Inferred from Reviewed Computational Analysis (RCA)

**Author statement evidence codes:** <br>
Traceable Author Statement (TAS) <br>
Non-traceable Author Statement (NAS)

**Curator statement evidence codes:** <br>
Inferred by Curator (IC) <br>
No biological Data available (ND)

**Electronic annotation evidence code:** <br>
Inferred from Electronic Annotation (IEA)

"""    
    display(Markdown(s))  
                    
        
def _write_go_hdf():

    with open(os.devnull, 'w') as null, redirect_stdout(null):

        if not os.path.exists(f'{cache_dir}/go-basic.h5'):

            # Get http://geneontology.org/ontology/go-basic.obo
            obo_fname = download_and_move_go_basic_obo(prt=null)

            # Download Associations, if necessary
            # Get ftp://ftp.ncbi.nlm.nih.gov/gene/DATA/gene2go.gz
            file_gene2go = download_ncbi_associations(prt=null)

            rows = [(e.id, e.name, e.defn) 
                    for e in OBOReader(obo_file=f'{cache_dir}/go-basic.obo', 
                                       optional_attrs=['def'])]
            df = pd.DataFrame().from_records(
                rows, columns=['goterm', 'goname', 'description'])
            df.to_hdf(f'{cache_dir}/go-basic.h5', key='df', format='table', 
                      data_columns=['goterm', 'goname'])


def go_term2name(term:str) -> str:
    """
    Converts a GO term to its name.

    Parameters
    ----------
    term : 
        GO term

    Returns
    -------
    :
        GO term name.
    """
    _write_go_hdf()
    with pd.HDFStore(f'{cache_dir}/go-basic.h5', 'r') as store:
        entry = store.select("df", "goterm == %r" % term).iloc[0]

    return entry.goterm


def go_name2term(name:str) -> str:
    """
    Converts a GO term name to its term.

    Parameters
    ----------
    name : 
        GO term name

    Returns
    -------
    :
        GO term.
    """
    _write_go_hdf()
    with pd.HDFStore(f'{cache_dir}/go-basic.h5', 'r') as store:
        entry = store.select("df", "goname == %r" % name.lower()).iloc[0]
    return entry.goterm


def go_info(terms:Union[str,List[str]]) -> None:
    """
    Displays HML formatted information about the given GO terms.

    Parameters
    ----------
    terms : 
        A GO term or list of GO terms to display information for.
    """

    if type(terms) is pd.core.series.Series:
        terms = terms.tolist()

    if type(terms) is not list:
        terms = [terms]

    _write_go_hdf()

    df = pd.read_hdf(f'{cache_dir}/go-basic.h5')
    for term in terms:
        entry = df.loc[df.goterm == term].iloc[0]
        desc = re.search(r'"([^"]+)"', entry.description).group(1)
        s = f'**<span style="color:gray;">{term}:</span>** ' + \
            f'**{entry.goname}**  \n {desc}    \n\n ----'        
        display(Markdown(s))


class WrSubObo(object):
    """
    Read a large GO-DAG from an obo file. Write a subset GO-DAG into
    a small obo file.
    """

    def __init__(self, fin_obo=None, optional_attrs=None, load_obsolete=None):
        self.fin_obo = fin_obo
        if fin_obo is not None:
            self.godag = GODag(fin_obo, optional_attrs, load_obsolete) 
        else:
            self.godag = None            
        self.relationships = \
            optional_attrs is not None and 'relationship' in optional_attrs

    def wrobo(self, fout_obo, goid_sources):
        """Write a subset obo file containing GO ID sources and their parents."""
        goids_all = self._get_goids_all(goid_sources)
        with open(fout_obo, 'w') as prt:
            self._prt_info(prt, goid_sources, goids_all)
            self.prt_goterms(self.fin_obo, goids_all, prt)

    @staticmethod
    def prt_goterms(fin_obo, goids, prt, b_prt=True):
        """Print the specified GO terms for GO IDs in arg."""
        b_trm = False
        with open(fin_obo) as ifstrm:
            for line in ifstrm:
                if not b_trm:
                    if line[:6] == "[Term]":
                        b_trm = True
                        b_prt = False
                    elif line[:6] == "[Typedef]":
                        b_prt = True
                else:
                    if line[:6] == 'id: GO':
                        b_trm = False
                        b_prt = line[4:14] in goids
                        if b_prt:
                            prt.write("[Term]\n")
                if b_prt:
                    prt.write(line)

    @staticmethod
    def get_goids(fin_obo, name):
        """
        Get GO IDs whose name matches given name.
        """
        goids = set()
        # pylint: disable=unsubscriptable-object
        goterm = None
        with open(fin_obo) as ifstrm:
            for line in ifstrm:
                if goterm is not None:
                    semi = line.find(':')
                    if semi != -1:
                        goterm[line[:semi]] = line[semi+2:].rstrip()
                    else:
                        if name in goterm['name']:
                            goids.add(goterm['id'])
                        goterm = None
                elif line[:6] == "[Term]":
                    goterm = {}
        return goids

    def _get_goids_all(self, go_sources):
        """
        Given GO ID sources and optionally the relationship attribute, 
        return all GO IDs.
        """
        go2obj_user = {}
        objrel = CurNHigher(self.relationships, self.godag)
        objrel.get_id2obj_cur_n_high(go2obj_user, go_sources)
        goids = set(go2obj_user)
        for goterm in go2obj_user.values():
            if goterm.alt_ids:
                goids.update(goterm.alt_ids)
        return goids

    def _prt_info(self, prt, goid_sources, goids_all):
        """
        Print information describing how this obo setset was created.
        """
        prt.write("! Contains {N} GO IDs. Created using {M} GO sources:\n".format(
            N=len(goids_all), M=len(goid_sources)))
        for goid in goid_sources:
            prt.write("!    {GO}\n".format(GO=str(self.godag.get(goid, ""))))
        prt.write("\n")


class My_GOEnrichemntRecord(GOEnrichmentRecord):

    def __str__(self):
        return f'<{self.GO}>'


def go_enrichment(gene_list:list, taxid:int=9606, background_chrom:str=None, 
                  background_genes:list=None, terms:list=None, 
                  list_study_genes:list=False, alpha:float=0.05) -> pd.DataFrame:
    """
    Runs a GO enrichment analysis.

    Parameters
    ----------
    gene_list : 
        List of gene symbols or NCBI gene ids.
    taxid : 
        NCBI taxonomy ID, 9606 (human) or 1758 (mouse), by default 9606.
    background_chrom : 
        Name of chromosome, by default None. Limits analysis to this named chromosome
    background_genes : 
        List of genes for use as background in GO enrichment analysis, 
        by default None
    terms : 
        List of GO terms for use as background in GO enrichment analysis,
        by default None
    list_study_genes : 
        Whether to include lists of genes responsible for enrichment for each 
        identified GO term, by default False
    alpha : 
        False discovery significance cut-off, by default 0.05

    Returns
    -------
    :
        pd.DataFrame with columns: 
        - namespace: (BP, MF, CC)
        - term_id: GO term
        - e/p: enrichment or depletion
        - pval_uncorr: uncorrected p-value
        - p_fdr_bh: Benjamini-Hochberg corrected p-value
        - ratio: ratio of study genes in GO term
        - bg_ratio: ratio of background genes in GO term
        - obj: GOEnrichmentRecord object

    Examples
    --------
    ```python
    gene_list = ['TP53', 'BRCA1', 'BRCA2', 'EGFR', 'KRAS', 'PTEN', 'CDH1', 
                 'ATM', 'CHEK2', 'PALB2']
    results = go_enrichment(gene_list, taxid=9606, alpha=0.05)
    show_go_dag_enrichment_results(results.obj)
    ```

    """
    if taxid not in [9606, 17580]:
        raise ValueError('Only human and mouse (tax id 9606 and 17580) are supported')
    
    if issubclass(type(gene_list), list):
        gene_list = list(gene_list)
    
    if type(gene_list) is pd.core.series.Series:
        gene_list = gene_list.tolist()
    if type(terms) is pd.core.series.Series:
        terms = terms.tolist()

    _assert_entrez_email()

    gene_list = list(gene_list)
    
    taxid = _tidy_taxid(taxid)

    ncbi_tsv = f'{cache_dir}/{taxid}_protein_genes.txt'
    if not os.path.exists(ncbi_tsv):
        fetch_background_genes(taxid)

    with open(os.devnull, 'w') as null, redirect_stdout(null):

        obo_fname = download_and_move_go_basic_obo(prt=null)

        file_gene2go = download_ncbi_associations(prt=null)

        obodag = GODag(f"{cache_dir}/go-basic.obo", 
                       optional_attrs=['relationship', 'def'], prt=null)

        # read NCBI's gene2go. Store annotations in a list of namedtuples
        objanno = Gene2GoReader(file_gene2go, taxids=[taxid])

        # get associations for each branch of the GO DAG (BP, MF, CC)
        ns2assoc = objanno.get_ns2assc()

        # limit go dag to a sub graph including only specified terms and their children
        if terms is not None:
            sub_obo_name = f'{cache_dir}/' + \
                str(hash(''.join(sorted(terms)).encode())) + '.obo'  
            wrsobo = WrSubObo(obo_fname, optional_attrs=['relationship', 'def'])
            wrsobo.wrobo(sub_obo_name, terms)    
            obodag = GODag(sub_obo_name, 
                           optional_attrs=['relationship', 'def'], prt=null)

        # load background gene set of all genes
        background_genes_file = f'{cache_dir}/{taxid}_protein_genes.txt'
        if not os.path.exists(background_genes_file):
            fetch_background_genes(taxid)

        # # load any custum subset
        if background_genes:
            if not all(type(x) is int for x in background_genes):
                if all(x.isnumeric() for x in background_genes):
                    background_genes = list(map(str, background_genes))
                else:
                    background_genes = _cached_symbol2ncbi(
                        background_genes, taxid=taxid)
            df = pd.read_csv(background_genes_file, sep='\t')
            no_suffix = os.path.splitext(background_genes_file)[0]
            background_genes_file = \
                f'{no_suffix}_{hash("".join(map(str, sorted(background_genes))))}.txt'            
            df.loc[df.GeneID.isin(background_genes)].to_csv(background_genes_file, 
                                                            sep='\t', index=False)

        # limit background gene set
        if background_chrom is not None:
            df = pd.read_csv(background_genes_file, sep='\t')
            background_genes_file = \
                f'{os.path.splitext(background_genes_file)[0]}_{background_chrom}.txt'            
            df.loc[df.chromosome == background_chrom].to_csv(
                background_genes_file, sep='\t', index=False)

        output_py = f'{cache_dir}/{taxid}_background.py'
        ncbi_tsv_to_py(background_genes_file, output_py, prt=null)

        background_genes = importlib.import_module(f'geneinfo.cache.{taxid}_background')
        # background_genes_name = output_py.replace('.py', '').replace('/', '.')
        # background_genes = importlib.import_module(background_genes_name, 'geneinfo.cache')
        importlib.reload(background_genes)
        GeneID2nt = background_genes.GENEID2NT

        if not all(type(x) is int for x in gene_list):
            gene_list = _cached_symbol2ncbi(gene_list, taxid=taxid)

        goeaobj = GOEnrichmentStudyNS(
                GeneID2nt, # List of mouse protein-coding genes
                ns2assoc, # geneid/GO associations
                obodag, # Ontologies
                propagate_counts = False,
                alpha = 0.05, # default significance cut-off
                methods=['fdr_bh'],
                pvalcalc='fisher_scipy_stats') 

        goea_results_all = goeaobj.run_study(gene_list)


        rows = []
        columns = ['namespace', 'term_id', 'e/p', 'pval_uncorr', 'p_fdr_bh', 
                'ratio', 'bg_ratio', 'obj']
        if list_study_genes:
            columns.append('study_genes')
        for ntd in goea_results_all:
            
            # Hack. Changes __class__ of all instances...
            ntd.__class__ = My_GOEnrichemntRecord

            row = [ntd.NS, ntd.GO, ntd.enrichment, ntd.p_uncorrected,
                        ntd.p_fdr_bh, 
                        ntd.ratio_in_study[0] / ntd.ratio_in_study[1],
                        ntd.ratio_in_pop[0] /  ntd.ratio_in_pop[1], ntd]

            if list_study_genes:
                row.append(_cached_ncbi2symbol(sorted(ntd.study_items)))
            rows.append(row)
        df = (pd.DataFrame()
            .from_records(rows, columns=columns)
            .sort_values(by=['p_fdr_bh', 'ratio'])
            .reset_index(drop=True)
        )
        return df.loc[df.p_fdr_bh < alpha]



def show_go_dag_enrichment_results(
        results:Union[List[GOEnrichmentRecord],pd.Series]) -> None:
    """
    Displays a GO enrichment analysis results.

    Parameters
    ----------
    results : 
        List or Series of GO result objejcts from `obj` column in the 
        `pd.DataFrame` returned by `go_enrichment()`.

    Examples
    --------
    ```python
    gene_list = ['TP53', 'BRCA1', 'BRCA2', 'EGFR', 'KRAS', 'PTEN', 
                 'CDH1', 'ATM', 'CHEK2', 'PALB2']
    results = go_enrichment(gene_list, taxid=9606, alpha=0.05)
    show_go_dag_enrichment_results(results.obj)
    ```
    """
    if type(results) is pd.core.series.Series:
        results = results.tolist()
    with open(os.devnull, 'w') as null, redirect_stdout(null):
        plot_results(f'{cache_dir}/plot.png', results)
    return display(Image(f'{cache_dir}/plot.png'))
